/*   ncbicoll.c
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*            National Center for Biotechnology Information (NCBI)
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government do not place any restriction on its use or reproduction.
*  We would, however, appreciate having the NCBI and the author cited in
*  any work or product based on this material
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
* ===========================================================================
*
* File Name:  ncbicoll.c
*
* Author:  Sergei Egorov
*
* Version Creation Date:   12/16/94
*
* $Revision: 5.1 $
*
* File Description: 
*     collection procedures
*
* Modifications:  
* --------------------------------------------------------------------------
* Date     Name        Description of modification
* -------  ----------  -----------------------------------------------------
*
* $Log: ncbicoll.c,v $
 * Revision 5.1  1996/12/03  21:48:33  vakatov
 * Adopted for 32-bit MS-Windows DLLs
 *
 * Revision 5.0  1996/05/28  13:18:57  ostell
 * Set to revision 5.0
 *
 * Revision 4.0  1995/07/26  13:46:50  ostell
 * force revision to 4.0
 *
 * Revision 1.6  1995/05/15  18:45:58  ostell
 * added Log line
 *
*
*
*
* ==========================================================================
*/

#include <ncbi.h>
#include <ncbicoll.h>

/*********************************************************************
 *
 *  error posting 
 *
 *********************************************************************/

static CharPtr int_msg = "in implementation\n%s (%d)";
static CharPtr bad_arg_msg = "illegal argument\n%s (%d)";
static CharPtr null_arg_msg = "NULL argument\n%s (%d)";
static CharPtr null_coll_msg = "NULL collection argument\n%s (%d)";
static CharPtr null_pos_msg = "NULL position argument\n%s (%d)";
static CharPtr no_mem_msg = "Run out of memory\n%s (%d)";

#define INTERNAL_ERROR() ErrPostEx(SEV_ERROR,0,1,int_msg,__FILE__,__LINE__)
#define BAD_ARG_ERROR() ErrPostEx(SEV_ERROR,E_Programmer,1,bad_arg_msg,__FILE__,__LINE__)
#define NULL_ARG_ERROR() ErrPostEx(SEV_ERROR,E_Programmer,1,null_arg_msg,__FILE__,__LINE__)
#define NULL_COLL_ERROR() ErrPostEx(SEV_ERROR,E_Programmer,1,null_coll_msg,__FILE__,__LINE__)
#define NULL_POS_ERROR() ErrPostEx(SEV_ERROR,E_Programmer,1,null_pos_msg,__FILE__,__LINE__)
#define NO_MEM_WARNING() ErrPostEx(SEV_WARNING,0,0,no_mem_msg,__FILE__,__LINE__)
/*
#define DEBUG_OUT(s) ErrPostEx(SEV_WARNING,0,0,s "\n%s (%d)",__FILE__,__LINE__)
*/
#define DEBUG_OUT(s)

/*********************************************************************
 *
 *  blkstore - segmented storage allocation and access procedures
 *
 *********************************************************************/

typedef BytePtr seg_ptr_t;
typedef seg_ptr_t PNTR segmap_ptr_t;

typedef void (*blk_proc_t) (BytePtr pblk);
typedef void (*bp_vp_proc_t) (BytePtr pblk, Pointer pdata);

typedef struct blkstore {
  Int4         blksize;    /* size of block in bytes: sizeof(my_type) */
  Int4         nblkspseg;  /* number of blocks per seg */
  Int4         nsegs;      /* number of allocated segments */
  Int4         smsize;     /* total # of entries in seg map (>= nsegs) */
  segmap_ptr_t smp;        /* pointer to seg map */
} blkstore_t, PNTR blkstore_ptr_t;

#define SEGMAP_DELTA  10     /* number of entries to add on realloc */
#define DEF_SEG_SIZE  KBYTE  /* default segment size in bytes */

/*
 *  constructors & destructors
 */

static Boolean blkstore_init (blkstore_ptr_t bsp, Int4 blksize,
                                 Int4 nblkspseg)
{
  Int4 blkbytes = blksize * nblkspseg;
  if (bsp != NULL && blksize > 0 && nblkspseg > 0 && blkbytes <= MAXALLOC) {
    size_t smsize = SEGMAP_DELTA * sizeof (seg_ptr_t);
    segmap_ptr_t smp = (segmap_ptr_t) MemNew (smsize);
    if (smp != NULL) {
      bsp->blksize = blksize;
      bsp->nblkspseg = nblkspseg;
      bsp->nsegs = 0;
      bsp->smsize = SEGMAP_DELTA;
      bsp->smp = smp;
      return TRUE;
    } else {
      NO_MEM_WARNING();
      MemFree (smp);
    }
  } else BAD_ARG_ERROR();
  return FALSE;
}

static void blkstore_fini (blkstore_ptr_t bsp)
{
  if (bsp != NULL) {
    segmap_ptr_t smp = bsp->smp;
    if (smp != NULL) {
      Int4 cnt, i;
      for (cnt = bsp->smsize, i = 0; i < cnt; i++) MemFree (smp[i]);
      MemFree (smp);
    }
  } else NULL_COLL_ERROR();
}

/*
 *  blkstore operations
*/

static Int4 blkstore_get_size (blkstore_ptr_t bsp)
{
  if (bsp != NULL) return bsp->nblkspseg * bsp->nsegs;
  NULL_COLL_ERROR(); return 0;
}

static Int4 blkstore_set_size (blkstore_ptr_t bsp, Int4 newsize)
{
  if (bsp != NULL) {
    Int4 nbps = bsp->nblkspseg;
    Int4 bsize = bsp->blksize;
    Int4 reqnsegs = (newsize + nbps - 1) / nbps;
    if (reqnsegs > bsp->nsegs) {
      DEBUG_OUT("blkstore:allocating extra segments");
      /* allocate extra segments */
      if (reqnsegs > bsp->smsize) {
        /* expand segmap if needed */
        Int4 newsmsize = reqnsegs + SEGMAP_DELTA;
        segmap_ptr_t smp = (segmap_ptr_t)
          MemNew ((size_t)(newsmsize * sizeof (seg_ptr_t)));
        /* if no room for segmap, just return current size */
        if (smp == NULL) {
          NO_MEM_WARNING();
          return bsp->nsegs * nbps;
        }
        MemCpy ((void *) smp, (const void *) bsp->smp, (size_t)(bsp->smsize * sizeof (seg_ptr_t)));
        MemFree (bsp->smp);
        bsp->smp = smp;
        bsp->smsize = newsmsize;
        DEBUG_OUT("blkstore:segmap expanded");
      }
      while (bsp->nsegs < reqnsegs) {
        seg_ptr_t sp = MemNew ((size_t)(bsize * nbps));
        if (sp == NULL) {
          NO_MEM_WARNING();
          break;
        }
        bsp->smp[bsp->nsegs] = sp;
        bsp->nsegs ++;
      }
    } else if (reqnsegs < bsp->nsegs) {
      /* free extra segments */
      while (bsp->nsegs > reqnsegs) {
        bsp->smp[bsp->nsegs] = MemFree (bsp->smp[bsp->nsegs]);
        bsp->nsegs --;
      }
    }
    return bsp->nsegs * nbps;
  } else NULL_COLL_ERROR();
  return 0;
}

static BytePtr blkstore_ref (blkstore_ptr_t bsp, Int4 index)
{
  if (bsp != NULL && index >= 0 && index < blkstore_get_size (bsp)) {
    Int4 segno = index / bsp->nblkspseg;
    Int4 offset = (index % bsp->nblkspseg) * bsp->blksize;
    return (BytePtr)(bsp->smp[segno]) + offset;
  } else BAD_ARG_ERROR();
  return NULL; /* null bsp or index out of range */
}

/* ToDo: use MemMove for contiguous sequences of blocks */
static void blkstore_move (blkstore_ptr_t bsp, Int4 idst, Int4 isrc,
                           Int4 nblks)
{
  if (bsp != NULL) {
    Int4 bsize = bsp->blksize;
    BytePtr src, dst;
    if (idst > isrc) {
      isrc += nblks;
      idst += nblks;
      while (nblks-- > 0) {
        --idst; --isrc;
        dst = blkstore_ref (bsp, idst);
        src = blkstore_ref (bsp, isrc);
        if (src == NULL || dst == NULL) break;
        MemCpy ((void *) dst, (const void *) src, (size_t)bsize);
      }
    } else {
      while (nblks-- > 0) {
        dst = blkstore_ref (bsp, idst);
        src = blkstore_ref (bsp, isrc);
        if (src == NULL || dst == NULL) break;
        MemCpy ((void *) dst, (const void *) src, (size_t)bsize);
        ++idst; ++isrc;
      }
    }
  } else NULL_COLL_ERROR();
}

static void blkstore_scan (blkstore_ptr_t bsp, Int4 ifrom, Int4 nblks,
                           bp_vp_proc_t scanproc, Pointer pdata)
{
  if (scanproc == NULL) return;
  if (bsp != NULL && ifrom >= 0) {
    segmap_ptr_t smp = bsp->smp;
    Int4 nblkspseg = bsp->nblkspseg;
    Int4 bssize = nblkspseg * bsp->nsegs;
    Int4 blksize = bsp->blksize;
    Int4 index;
    for (index = ifrom; index < bssize && nblks > 0; --nblks, ++index) {
      Int4 segno = index / nblkspseg;
      Int4 offset = (index % nblkspseg) * blksize;
      scanproc ((BytePtr)(smp[segno]) + offset, pdata);
    }
  } else NULL_ARG_ERROR();
}

static void blkstore_apply (blkstore_ptr_t bsp, Int4 ifrom, Int4 nblks,
                            blk_proc_t proc)
{
  if (proc == NULL) return;
  if (bsp != NULL && ifrom >= 0) {
    segmap_ptr_t smp = bsp->smp;
    Int4 nblkspseg = bsp->nblkspseg;
    Int4 bssize = nblkspseg * bsp->nsegs;
    Int4 blksize = bsp->blksize;
    Int4 index;
    for (index = ifrom; index < bssize && nblks > 0; --nblks, ++index) {
      Int4 segno = index / nblkspseg;
      Int4 offset = (index % nblkspseg) * blksize;
      proc ((BytePtr)(smp[segno]) + offset);
    }
  } else BAD_ARG_ERROR();
}


/*********************************************************************
 *
 *  Variant - a(nother) universal data type
 *
 *********************************************************************/

NLM_EXTERN Variant LIBCALL Nlm_PtrToVnt (Pointer p)
{
  Variant v; v.ptrvalue = p; return v;
}

NLM_EXTERN Variant LIBCALL Nlm_HandToVnt (Handle h)
{
  Variant v; v.handvalue = h; return v;
}

NLM_EXTERN Variant LIBCALL Nlm_IntToVnt (Int4 i)
{
  Variant v; v.intvalue = i; return v;
}

NLM_EXTERN Variant LIBCALL Nlm_FloatToVnt (FloatHi f)
{
  Variant v; v.realvalue = f; return v;
}

NLM_EXTERN Variant LIBCALL Nlm_FnPtrToVnt (FnPtr f)
{
  Variant v; v.funcvalue = f; return v;
}

NLM_EXTERN void Nlm_VntFreePtr (VariantPtr vp)
{
  if (vp != NULL) MemFree (vp->ptrvalue);
  else NULL_ARG_ERROR();
}

NLM_EXTERN void Nlm_VntFreeHand (VariantPtr vp)
{
  if (vp != NULL) HandFree (vp->handvalue);
  else NULL_ARG_ERROR();
}

NLM_EXTERN Boolean Nlm_VntTestEqToPtr (VariantPtr vp, VoidPtr ptr)
{
  if (vp != NULL) return vp->ptrvalue == ptr;
  NULL_ARG_ERROR(); return FALSE;
}

NLM_EXTERN Boolean Nlm_VntTestEqToHand (VariantPtr vp, Handle PNTR phand)
{
  if (vp != NULL && phand != NULL) return vp->handvalue == *phand;
  NULL_ARG_ERROR(); return FALSE;
}

NLM_EXTERN Boolean Nlm_VntTestEqToInt4 (VariantPtr vp, Int4Ptr pint)
{
  if (vp != NULL && pint != NULL) return vp->intvalue == *pint;
  NULL_ARG_ERROR(); return FALSE;
}


NLM_EXTERN Int2 Nlm_VntCmpStrings (VariantPtr vp1, VariantPtr vp2)
{
  if (vp1 != NULL && vp2 != NULL)
    return StrCmp ((CharPtr)vp1->ptrvalue, (CharPtr)vp2->ptrvalue);
  NULL_ARG_ERROR(); return 0;
}

NLM_EXTERN Int2 Nlm_VntCmpIStrings (VariantPtr vp1, VariantPtr vp2)
{
  if (vp1 != NULL && vp2 != NULL)
    return StrICmp ((CharPtr)vp1->ptrvalue, (CharPtr)vp2->ptrvalue);
  NULL_ARG_ERROR(); return 0;
}


/*********************************************************************
 *
 *  VntArray - dynamic array of Variants
 *
 *********************************************************************/

typedef struct varray {
  blkstore_t   store;       /* segmented storage of variants */
  Int4         delta;       /* min number of nodes to add on realloc */
  Int4         length;      /* current length of the array */
} varray_t, PNTR varray_ptr_t;

/*
 *  varray constructors & destructors
 */

static Boolean varray_init (varray_ptr_t vap, Int4 delta)
{
  if (vap != NULL) {
    /* calculate reasonable BlockStore segment size */
    Int4 blksize = sizeof (Variant);
    Int4 nblkspseg = DEF_SEG_SIZE / blksize; /* default is 1K */
    Int4 blkbytes = blksize * delta;
    if (blkbytes > 0 && blkbytes <= MAXALLOC) nblkspseg = delta;
    else delta = nblkspseg;
    /* init BlockStore and array-specific fields */
    if (blkstore_init (&vap->store, blksize, nblkspseg)) {
      vap->delta = delta;
      vap->length = 0;
      return TRUE;
    }
  } else NULL_COLL_ERROR();
  return FALSE;
}

static varray_ptr_t varray_new (Int4 delta)
{
  varray_ptr_t vap = MemNew (sizeof (varray_t));
  if (vap != NULL) {
    if (! varray_init (vap, delta))
      vap = MemFree (vap);
  } else NO_MEM_WARNING();
  return vap;
}

static void varray_fini (varray_ptr_t vap, VntProc fini)
{
  if (vap != NULL) {
    blkstore_apply (&vap->store, 0, vap->length, (blk_proc_t) fini);
    blkstore_fini (&vap->store);
  } else NULL_COLL_ERROR();
}

static void varray_free (varray_ptr_t vap, VntProc fini)
{
  varray_fini (vap, fini);
  MemFree (vap);
}

static void varray_set_len (varray_ptr_t vap, Int4 len, 
                            VntProc init, VntProc fini)
{
  if (len >= 0 && vap != NULL) {
    Int4 newlen;
    if (len > vap->length) {
      newlen = blkstore_set_size (&vap->store, len);
      if (newlen < vap->length) INTERNAL_ERROR();
      else if (newlen < len) NO_MEM_WARNING();
      else {
        blkstore_apply (&vap->store, vap->length, len - vap->length, 
                        (blk_proc_t) init);
        vap->length = len;
      }
    } else if (len < vap->length) {
      blkstore_apply (&vap->store, len, vap->length - len, 
                      (blk_proc_t) fini);
      newlen = blkstore_set_size (&vap->store, len);
      if (newlen < len) INTERNAL_ERROR();
      else {
        vap->length = len;
      }
    }
  } else BAD_ARG_ERROR();
}

static VariantPtr varray_ref (varray_ptr_t vap, Int4 index, 
                              Boolean grow, VntProc init)
{
  if (index >= 0 && vap != NULL) {
    if (index >= vap->length) {
      if (grow) varray_set_len (vap, index + 1, init, NULL);
      else BAD_ARG_ERROR();
    }
    if (index < vap->length) {
      return ((VariantPtr) blkstore_ref (&vap->store, index));
    } /* else MO_MEM has been reported by varray_set_len */
  } else BAD_ARG_ERROR();
  return NULL;
}

static void varray_insert (varray_ptr_t vap, Int4 index, Variant v, Int4 cnt)
{
  if (vap != NULL) {
    Int4 len = vap->length;
    if (index >= 0 && index < len && cnt >= 0) {
      varray_set_len (vap, len + cnt, NULL, NULL);
      if (vap->length == len + cnt) {
        blkstore_move (&vap->store, index + cnt, index, len - index);
        for (; cnt > 0; ++ index, -- cnt) {
          VariantPtr vp = (VariantPtr) blkstore_ref (&vap->store, index);
          if (vp == NULL) { INTERNAL_ERROR(); return; }
          *vp = v; 
        }
      }
    } else BAD_ARG_ERROR();
  } else NULL_COLL_ERROR();
}

static void varray_delete (varray_ptr_t vap, Int4 index, 
                           VntProc fini, Int4 cnt)
{
  if (vap != NULL) {
    Int4 len = vap->length;
    if (index >= 0 && index < len && cnt >= 0) {
      if (cnt > len - index) cnt = len - index;
      blkstore_apply (&vap->store, index, cnt, (blk_proc_t) fini);
      blkstore_move (&vap->store, index, index + cnt, len - index - cnt);
      varray_set_len (vap, len - cnt, NULL, NULL);
    } else BAD_ARG_ERROR();
  } else NULL_COLL_ERROR();
}


NLM_EXTERN VntArrayPtr LIBCALL Nlm_VntArrayNew (Int2 delta)
{ 
  return (VntArrayPtr) varray_new (delta);
}

NLM_EXTERN VntArrayPtr LIBCALL Nlm_VntArrayFree (VntArrayPtr vap, VntProc fini)
{
  varray_free ((varray_ptr_t) vap, fini);
  return NULL;
}

NLM_EXTERN Int4 LIBCALL Nlm_VntArrayLen (VntArrayPtr vap)
{
  return (vap == NULL) ? 0 : ((varray_ptr_t) vap)->length;
}

NLM_EXTERN Boolean LIBCALL Nlm_VntArrayIsEmpty (VntArrayPtr vap)
{
  return (vap == NULL) ? TRUE : (((varray_ptr_t) vap)->length == 0);
}

NLM_EXTERN void LIBCALL Nlm_VntArraySetLen (VntArrayPtr vap, Int4 len, 
                             VntProc init, VntProc fini)
{
  varray_set_len ((varray_ptr_t) vap, len, init, fini);
}

NLM_EXTERN VariantPtr LIBCALL Nlm_VntArrayRefAt (VntArrayPtr vap, Int4 index)
{
  return varray_ref ((varray_ptr_t) vap, index, FALSE, NULL);
}

NLM_EXTERN Variant LIBCALL Nlm_VntArrayGetAt (VntArrayPtr vap, Int4 index)
{
  VariantPtr vp = varray_ref ((varray_ptr_t) vap, index, FALSE, NULL);
  if (vp != NULL) return *vp;
  else return PtrToVnt (NULL);
}

NLM_EXTERN void LIBCALL Nlm_VntArraySetAt (VntArrayPtr vap, Int4 index, Variant v)
{
  VariantPtr vp = varray_ref ((varray_ptr_t) vap, index, FALSE, NULL);
  if (vp != NULL) *vp = v;
}

NLM_EXTERN VariantPtr LIBCALL Nlm_VntArrayRefAtGrow (VntArrayPtr vap, Int4 index, 
                                      VntProc init)
{
  return varray_ref ((varray_ptr_t) vap, index, TRUE, init);
}

NLM_EXTERN Variant LIBCALL Nlm_VntArrayGetAtGrow (VntArrayPtr vap, Int4 index, 
                                   VntProc init)
{
  VariantPtr vp = varray_ref ((varray_ptr_t) vap, index, TRUE, init);
  if (vp != NULL) return *vp;
  else return PtrToVnt (NULL);
}

NLM_EXTERN void LIBCALL Nlm_VntArraySetAtGrow (VntArrayPtr vap, Int4 index, Variant v, 
                                VntProc init)
{
  VariantPtr vp = varray_ref ((varray_ptr_t) vap, index, TRUE, init);
  if (vp != NULL) *vp = v;
}

NLM_EXTERN Int4 LIBCALL Nlm_VntArrayAdd (VntArrayPtr vap, Variant v)
{
  if (vap != NULL) {
    varray_ptr_t ap = (varray_ptr_t) vap;
    Int4 index = ap->length;
    VariantPtr vp = varray_ref (ap, index, TRUE, NULL);
    if (vp != NULL) {
      *vp = v; 
      return index;
    }
  } else NULL_COLL_ERROR();
  return -1;
}

NLM_EXTERN void LIBCALL Nlm_VntArrayInsertAt (VntArrayPtr vap, Int4 index, Variant v)
{
  varray_insert ((varray_ptr_t) vap, index, v, 1);
}

NLM_EXTERN Variant LIBCALL Nlm_VntArrayRemoveAt (VntArrayPtr vap, Int4 index)
{
  VariantPtr vp = varray_ref ((varray_ptr_t) vap, index, FALSE, NULL);
  if (vp != NULL) { 
    Variant v;
    v = *vp;
    varray_delete ((varray_ptr_t) vap, index, NULL, 1);
    return v;
  }
  else return PtrToVnt (NULL);
}

NLM_EXTERN void LIBCALL Nlm_VntArrayDeleteAt (VntArrayPtr vap, Int4 index, VntProc fini)
{
  varray_delete ((varray_ptr_t) vap, index, fini, 1);
}

NLM_EXTERN void LIBCALL Nlm_VntArrayDeleteAll (VntArrayPtr vap, VntProc fini)
{
  varray_set_len ((varray_ptr_t) vap, 0, NULL, fini);
}


/*********************************************************************
 *
 *  vnode(1,2) - double link variant nodes
 *
 *********************************************************************/

typedef struct vnode1 {
  struct vnode1 PNTR pnext;
  struct vnode1 PNTR pprev;
  Variant variants[1];
} vnode1_t, PNTR vnode_ptr_t, PNTR PNTR vnode_ptrptr_t;

typedef struct vnode2 {
  struct vnode1 PNTR pnext;
  struct vnode1 PNTR pprev;
  Variant variants[2];
} vnode2_t;

/*
 * Insert the node pointed by vnp before/after dst; if dst is NULL,
 * before NULL means "at tail", after NULL means "at head".
 * ppfirst and pplast are addresses of pointers to the ends of the list.
 * Note: for correct operation at least one of them should be non-NULL.
 */
static void vnode_insert (vnode_ptr_t vnp, Boolean before, vnode_ptr_t dst,
                          vnode_ptrptr_t ppfirst, vnode_ptrptr_t pplast)
{
  if (vnp != NULL) {
    vnode_ptr_t pfirst = (ppfirst != NULL) ? *ppfirst : NULL;
    vnode_ptr_t plast = (pplast != NULL) ? *pplast : NULL;
    vnode_ptr_t pprev = before ? ((dst != NULL) ? dst->pprev : plast) : dst;
    vnode_ptr_t pnext = before ? dst : ((dst != NULL) ? dst->pnext : pfirst);
    vnp->pprev = pprev;
    if (pprev != NULL) pprev->pnext = vnp;
    else if (ppfirst != NULL) *ppfirst = vnp;
    vnp->pnext = pnext;
    if (pnext != NULL) pnext->pprev = vnp;
    else if (pplast != NULL) *pplast = vnp;
  } else NULL_POS_ERROR();
}

/*
 * Remove the node pointed by vnp from the list; vnp is not changed
 * and both fields of it could be used after removal.
 * ppfirst and pplast are addresses of pointers to the ends of the list.
 * Note: for correct operation at least one of them should be non-NULL.
 */
static void vnode_remove (vnode_ptr_t vnp,
                          vnode_ptrptr_t ppfirst, vnode_ptrptr_t pplast)
{
  if (vnp != NULL) {
    vnode_ptr_t pprev = vnp->pprev;
    vnode_ptr_t pnext = vnp->pnext;
    if (pprev != NULL) pprev->pnext = pnext;
    else if (ppfirst != NULL) *ppfirst = pnext;
    if (pnext != NULL) pnext->pprev = pprev;
    else if (pplast != NULL) *pplast = pprev;
  } else NULL_POS_ERROR();
}


/*********************************************************************
 *
 *  vlist - doubly-linked list of Variant[1,2]
 *
 *********************************************************************/

typedef struct vlist {
  blkstore_t   store;       /* segmented storage of vnodes */
  Int2         nvnts;       /* number of variants in each node (>= 1) */
  Int4         nstored;     /* number of stored (free+used) nodes */
  Int4         delta;       /* min number of nodes to add on realloc */
  Int4         length;      /* current length of the list */
  vnode_ptr_t  pfirst;      /* ptr to the first node or NULL */
  vnode_ptr_t  plast;       /* ptr to the last node or NULL */
  vnode_ptr_t  pfree;       /* ptr to the first free node or NULL */
} vlist_t, PNTR vlist_ptr_t;

/*
 *  vlist constructors & destructors
 */

static Boolean vlist_init (vlist_ptr_t vlp, Int2 nvnts, Int4 delta)
{
  if (vlp != NULL) {
    /* calculate reasonable BlockStore segment size */
    Int4 blksize = (nvnts == 1) ? sizeof (vnode1_t) : sizeof (vnode2_t);
    Int4 nblkspseg = DEF_SEG_SIZE / blksize; /* default is 1K */
    Int4 blkbytes = blksize * delta;
    if (blkbytes > 0 && blkbytes <= MAXALLOC) nblkspseg = delta;
    else delta = nblkspseg;
    /* init BlockStore and array-specific fields */
    if (blkstore_init (&vlp->store, blksize, nblkspseg)) {
      vlp->nstored = 0;
      vlp->nvnts = nvnts;
      vlp->delta = delta;
      vlp->nstored = 0;
      vlp->length = 0;
      vlp->pfirst = NULL;
      vlp->plast = NULL;
      vlp->pfree = NULL;
      return TRUE;
    }
  } else NULL_COLL_ERROR();
  return FALSE;
}

static vlist_ptr_t vlist1_new (Int4 delta)
{
  vlist_ptr_t vlp = MemNew (sizeof (vlist_t));
  if (vlp != NULL) {
    if (! vlist_init (vlp, 1, delta))
      vlp = MemFree (vlp);
  } else NO_MEM_WARNING();
  return vlp;
}

static vlist_ptr_t vlist2_new (Int4 delta)
{
  vlist_ptr_t vlp = MemNew (sizeof (vlist_t));
  if (vlp != NULL) {
    if (! vlist_init (vlp, 2, delta))
      vlp = MemFree (vlp);
  } else NO_MEM_WARNING();
  return vlp;
}

static void vlist1_fini (vlist_ptr_t vlp, VntProc fini)
{
  if (vlp != NULL) {
    if (fini != NULL) {
      /* finalize all Variants in a list */
      vnode_ptr_t vnp;
      for (vnp = vlp->pfirst; vnp != NULL; vnp = vnp->pnext) {
        fini (vnp->variants);
      }
    }
    blkstore_fini (&vlp->store);
  } else NULL_COLL_ERROR();
}

static void vlist2_fini (vlist_ptr_t vlp, VntProc finikey, VntProc finival)
{
  if (vlp != NULL) {
    /* finalize all Variants in a list */
    vnode_ptr_t vnp;
    for (vnp = vlp->pfirst; vnp != NULL; vnp = vnp->pnext) {
      if (finikey != NULL) finikey (&vnp->variants[0]);
      if (finival != NULL) finival (&vnp->variants[1]);
    }
    blkstore_fini (&vlp->store);
  } else NULL_COLL_ERROR();
}

static void vlist1_free (vlist_ptr_t vlp, VntProc fini)
{
  vlist1_fini (vlp, fini);
  MemFree (vlp);
}

static void vlist2_free (vlist_ptr_t vlp, VntProc finikey, VntProc finival)
{
  vlist2_fini (vlp, finikey, finival);
  MemFree (vlp);
}

/*
 *  vlist Operations
 */

static Int4 vlist_length (vlist_ptr_t vlp)
{
  return (vlp != NULL) ? vlp->length : 0;
}

static vnode_ptr_t vlist_first (vlist_ptr_t vlp)
{
  return (vlp != NULL) ? vlp->pfirst : NULL;
}

static vnode_ptr_t vlist_last (vlist_ptr_t vlp)
{
  return (vlp != NULL) ? vlp->plast : NULL;
}

/* Note: free list is linked thru pnext fields only */

static void free_vnode (vnode_ptr_t vnp, vlist_ptr_t vlp)
{
  if (vnp != NULL && vlp != NULL) {
    vnp->pnext = vlp->pfree;
    vlp->pfree = vnp;
  } else NULL_ARG_ERROR();
}

static vnode_ptr_t vlist_insert_new (vlist_ptr_t vlp, Boolean before,
                                     vnode_ptr_t dst)
{
  if (vlp != NULL) {
    vnode_ptr_t vnp = NULL;
    /* get free link to vnp; allocate memory if necessary */
    if (vlp->pfree == NULL) {
      Int4 bssize = blkstore_get_size (&vlp->store);
      if (bssize <= vlp->length) /* cannot be < ! */
        bssize = blkstore_set_size (&vlp->store, bssize + vlp->delta);
      if (bssize > vlp->length)
        blkstore_scan (&vlp->store, vlp->length, bssize - vlp->length,
                       (bp_vp_proc_t) free_vnode, vlp);
    }
    if (vlp->pfree != NULL) {
      vnp = vlp->pfree;
      vlp->pfree = vnp->pnext;
    } else NO_MEM_WARNING();
    /* insert vnp */
    if (vnp != NULL) {
      vnode_insert (vnp, before, dst, &vlp->pfirst, &vlp->plast);
      vlp->length ++;
      return vnp;
    } 
  } else NULL_COLL_ERROR();
  return NULL;
}

static void vlist_remove (vlist_ptr_t vlp, vnode_ptr_t vnp)
{
  if (vlp != NULL && vnp != NULL) {
    vnode_remove (vnp, &vlp->pfirst, &vlp->plast);
    free_vnode (vnp, vlp);
    vlp->length --;
  } else NULL_ARG_ERROR();
}


/*********************************************************************
 *
 *  VntList - doubly-linked list of Variants
 *
 *********************************************************************/

NLM_EXTERN VntListPtr LIBCALL Nlm_VntListNew (Int2 delta)
{
  return (VntListPtr) vlist1_new (delta);
}

NLM_EXTERN VntListPtr LIBCALL Nlm_VntListFree (VntListPtr vlp, VntProc fini)
{
  vlist1_free ((vlist_ptr_t) vlp, fini);
  return NULL;
}

NLM_EXTERN Int4 LIBCALL Nlm_VntListLen (VntListPtr vlp)
{
  return vlist_length ((vlist_ptr_t) vlp);
}

NLM_EXTERN Boolean LIBCALL Nlm_VntListIsEmpty (VntListPtr vlp)
{
  return vlist_length ((vlist_ptr_t) vlp) == 0;
}

NLM_EXTERN VntPos LIBCALL Nlm_VntListHeadPos (VntListPtr vlp)
{
  return (VntPos) vlist_first ((vlist_ptr_t) vlp);
}

NLM_EXTERN VntPos LIBCALL Nlm_VntListTailPos (VntListPtr vlp)
{
  return (VntPos) vlist_last ((vlist_ptr_t) vlp);
}

NLM_EXTERN VntPos LIBCALL Nlm_VntListPrevPos (VntListPtr vlp, VntPos vnp)
{
  return (vnp == NULL) ? NULL : (VntPos)(((vnode_ptr_t)vnp)->pprev);
}

NLM_EXTERN VntPos LIBCALL Nlm_VntListNextPos (VntListPtr vlp, VntPos vnp)
{
  return (vnp == NULL) ? NULL : (VntPos)(((vnode_ptr_t)vnp)->pnext);
}

NLM_EXTERN Variant LIBCALL Nlm_VntListGetAt (VntListPtr vlp, VntPos vnp)
{
  if (vlp == NULL || vnp == NULL) {
    NULL_ARG_ERROR(); return PtrToVnt (NULL);
  }
  return ((vnode_ptr_t)vnp)->variants[0];
}

NLM_EXTERN Variant LIBCALL Nlm_VntListGetNext (VntListPtr vlp, VntPosPtr vnpp)
{
  vnode_ptr_t np;
  if (vlp == NULL || vnpp == NULL ||
      (np = (vnode_ptr_t) *vnpp) == NULL) {
    NULL_ARG_ERROR(); return PtrToVnt (NULL);
  }
  *vnpp = (VntPos)(np->pnext);
  return np->variants[0];
}

NLM_EXTERN Variant LIBCALL Nlm_VntListGetPrev (VntListPtr vlp, VntPosPtr vnpp)
{
  vnode_ptr_t np;
  if (vlp == NULL || vnpp == NULL ||
      (np = (vnode_ptr_t) *vnpp) == NULL) {
    NULL_ARG_ERROR(); return PtrToVnt (NULL);
  }
  *vnpp = (VntPos)(np->pprev);
  return np->variants[0];
}

NLM_EXTERN void LIBCALL Nlm_VntListSetAt (VntListPtr vlp, VntPos vnp, Variant v)
{
  vnode_ptr_t np;
  if (vlp == NULL || (np = (vnode_ptr_t) vnp) == NULL) {
    NULL_ARG_ERROR(); return;
  }
  np->variants[0] = v;
}

NLM_EXTERN VntPos LIBCALL Nlm_VntListInsertBefore (VntListPtr vlp, VntPos vnp, Variant v)
{
  vnode_ptr_t np = vlist_insert_new ((vlist_ptr_t) vlp, TRUE,
                                     (vnode_ptr_t) vnp);
  if (np != NULL) np->variants[0] = v;
  else NO_MEM_WARNING();
  return (VntPos) np;
}

NLM_EXTERN VntPos LIBCALL Nlm_VntListInsertAfter (VntListPtr vlp, VntPos vnp, Variant v)
{
  vnode_ptr_t np = vlist_insert_new ((vlist_ptr_t) vlp, FALSE,
                                     (vnode_ptr_t) vnp);
  if (np != NULL) np->variants[0] = v;
  else NO_MEM_WARNING();
  return (VntPos) np;
}

NLM_EXTERN Variant LIBCALL Nlm_VntListRemoveAt (VntListPtr vlp, VntPos vnp)
{
  if (vlp != NULL && vnp != NULL) {
    vnode_ptr_t np;
    Variant v;
    np = (vnode_ptr_t)vnp;
    v = np->variants[0];
    vlist_remove ((vlist_ptr_t) vlp, np);
    return v;
  } else {
    NULL_ARG_ERROR(); return PtrToVnt (NULL);
  }
}

NLM_EXTERN void LIBCALL Nlm_VntListDeleteAt (VntListPtr vlp, VntPos vnp, VntProc fini)
{
  if (vlp != NULL && vnp != NULL) {
    vnode_ptr_t np = (vnode_ptr_t)vnp;
    if (fini != NULL) fini (&np->variants[0]);
    vlist_remove ((vlist_ptr_t) vlp, np);
  } else NULL_ARG_ERROR();
}

NLM_EXTERN void LIBCALL Nlm_VntListDeleteAll (VntListPtr vlp, VntProc fini)
{
  if (vlp != NULL) {
    vlist_ptr_t lp = (vlist_ptr_t) vlp;
    vnode_ptr_t np = lp->pfirst;
    while (np != NULL) {
      vnode_ptr_t pnext = np->pnext;
      if (fini != NULL) fini (&np->variants[0]);
      free_vnode (np, lp);
      np = pnext;
    }
    lp->pfirst = lp->plast = NULL;
    lp->length = 0;
  } else NULL_COLL_ERROR();
}

NLM_EXTERN VntPos LIBCALL Nlm_VntListIndexPos (VntListPtr vlp, Int4 index)
{
  if (vlp != NULL) {
    if (index < 0) {
      return NULL;
    } else {
      vlist_ptr_t lp = (vlist_ptr_t) vlp;
      vnode_ptr_t np = lp->pfirst;
      while (np != NULL && index > 0) {
        np = np->pnext;
        index --;
      }
      return (VntPos) np; 
    }
  } else NULL_COLL_ERROR();
  return NULL;
}

NLM_EXTERN VntPos LIBCALL Nlm_VntListFindNextPos (VntListPtr vlp, VntPos vnp, 
                                   VntTestProc tp, VoidPtr pdata)
{
  if (vlp != NULL && tp != NULL) {
    vlist_ptr_t lp = (vlist_ptr_t) vlp;
    vnode_ptr_t np = (vnp == NULL) ? lp->pfirst : (vnode_ptr_t) vnp;
    while (np != NULL) {
      if (tp (&np->variants[0], pdata))
        return (VntPos) np;
      np = np->pnext;
    }
    return NULL; /* not found */
  } else NULL_ARG_ERROR();
  return NULL;
}

NLM_EXTERN VntPos LIBCALL Nlm_VntListFindPrevPos (VntListPtr vlp, VntPos vnp, 
                                   VntTestProc tp, VoidPtr pdata)
{
  if (vlp != NULL && tp != NULL) {
    vlist_ptr_t lp = (vlist_ptr_t) vlp;
    vnode_ptr_t np = (vnp == NULL) ? lp->plast : (vnode_ptr_t) vnp;
    while (np != NULL) {
      if (tp (&np->variants[0], pdata))
        return (VntPos) np;
      np = np->pprev;
    }
    return NULL; /* not found */
  } else NULL_ARG_ERROR();
  return NULL;
}

NLM_EXTERN VntPos LIBCALL Nlm_VntSortedListLookupPos (VntListPtr vlp, Variant v, 
                                       VntCmpProc cmp)
{
  if (vlp != NULL && cmp != NULL) {
    vlist_ptr_t lp = (vlist_ptr_t) vlp;
    vnode_ptr_t np;
    for (np = lp->pfirst; np != NULL; np = np->pnext) {
      Int2 ord = cmp (&v, &np->variants[0]);
      if (ord < 0) break;
      if (ord == 0) return (VntPos) np;
    }
    return NULL; /* not found */
  } else NULL_ARG_ERROR();
  return NULL;
}

NLM_EXTERN VntPos LIBCALL Nlm_VntSortedListInsert (VntListPtr vlp, Variant v, 
                                    VntCmpProc cmp)
{
  if (vlp != NULL && cmp != NULL) {
    vlist_ptr_t lp = (vlist_ptr_t) vlp;
    vnode_ptr_t np;
    for (np = lp->pfirst; np != NULL; np = np->pnext) {
      if (cmp (&v, &np->variants[0]) <= 0) break;
    }
    np = vlist_insert_new (lp, TRUE, np);  /* insert before */
    if (np != NULL) {
      np->variants[0] = v;
      return (VntPos) np;
    } else NO_MEM_WARNING();
  } else NULL_ARG_ERROR();
  return NULL;
}


/*********************************************************************
 *
 *  VntMap - maps Variants to Variants
 *
 *********************************************************************/

NLM_EXTERN VntMapPtr LIBCALL Nlm_VntMapNew (Int2 delta)
{
  return (VntMapPtr) vlist2_new (delta);
}

NLM_EXTERN VntMapPtr LIBCALL Nlm_VntMapFree (VntMapPtr vmp, VntProc finikey, 
                              VntProc finival)
{
  vlist2_free ((vlist_ptr_t) vmp, finikey, finival);
  return NULL;
}

NLM_EXTERN Int4 LIBCALL Nlm_VntMapLen (VntMapPtr vmp)
{
  return vlist_length ((vlist_ptr_t) vmp);
}

NLM_EXTERN Boolean LIBCALL Nlm_VntMapIsEmpty (VntMapPtr vmp)
{
  return vlist_length ((vlist_ptr_t) vmp) == 0;
}

NLM_EXTERN VntPos LIBCALL Nlm_VntMapStartPos (VntMapPtr vmp)
{
  return (VntPos) vlist_first ((vlist_ptr_t) vmp);
}

NLM_EXTERN VntPos LIBCALL Nlm_VntMapNextPos (VntMapPtr vmp, VntPos vnp)
{
  return (vnp == NULL) ? NULL : (VntPos)(((vnode_ptr_t)vnp)->pnext);
}

NLM_EXTERN Variant LIBCALL Nlm_VntMapGetKeyAt (VntMapPtr vmp, VntPos vnp)
{
  if (vmp == NULL || vnp == NULL) {
    NULL_ARG_ERROR(); return PtrToVnt (NULL);
  }
  return ((vnode_ptr_t)vnp)->variants[0];
}

NLM_EXTERN Variant LIBCALL Nlm_VntMapGetValAt (VntMapPtr vmp, VntPos vnp)
{
  Variant PNTR variPtr;

  if (vmp == NULL || vnp == NULL) {
    NULL_ARG_ERROR(); return PtrToVnt (NULL);
  }
  variPtr = &(((vnode_ptr_t)vnp)->variants[0]);
  return variPtr[1];
}

NLM_EXTERN Variant LIBCALL Nlm_VntMapReplaceValAt (VntMapPtr vmp, VntPos vnp, Variant newv)
{
  if (vmp != NULL || vnp != NULL) {
    vnode_ptr_t np;
    Variant v;
    Variant PNTR variPtr;
    np = (vnode_ptr_t) vnp;
    variPtr = &(np->variants[0]);
    v = variPtr[1];
    variPtr[1] = newv;
    return v;
  } else NULL_ARG_ERROR();
  return PtrToVnt (NULL); 
}

NLM_EXTERN void LIBCALL Nlm_VntMapGetNextAssoc (VntMapPtr vmp, VntPosPtr vnpp,
                                            VariantPtr pkey, VariantPtr pval)
{
  vnode_ptr_t np;
  Variant PNTR variPtr;
  if (vmp != NULL && vnpp != NULL && (np = (vnode_ptr_t) *vnpp) != NULL) {
    *vnpp = (VntPos)(np->pnext);
    variPtr = &(np->variants[0]);
    if (pkey != NULL) *pkey = variPtr[0];
    if (pval != NULL) *pval = variPtr[1];
  } else NULL_ARG_ERROR();
}

NLM_EXTERN VntPos LIBCALL Nlm_VntMapLookupPos (VntMapPtr vmp, Variant key, 
                                VntCmpProc cmpkeys)
{
  if (vmp != NULL && cmpkeys != NULL) {
    vlist_ptr_t lp = (vlist_ptr_t) vmp;
    vnode_ptr_t np;
    for (np = lp->pfirst; np != NULL; np = np->pnext) {
      Int2 ord = cmpkeys (&key, &np->variants[0]);
      if (ord < 0) break;
      if (ord == 0) return (VntPos) np;
    }
    return NULL; /* not found */
  } else NULL_ARG_ERROR();
  return NULL;
}

NLM_EXTERN Boolean LIBCALL Nlm_VntMapLookup (VntMapPtr vmp, Variant key, 
                                         VntCmpProc cmpkeys, VariantPtr pval)
{
  if (vmp != NULL && cmpkeys != NULL) {
    vlist_ptr_t lp = (vlist_ptr_t) vmp;
    vnode_ptr_t np;
    for (np = lp->pfirst; np != NULL; np = np->pnext) {
      Int2 ord = cmpkeys (&key, &np->variants[0]);
      if (ord < 0) break;
      if (ord == 0) {
        if (pval != NULL) {
          Variant PNTR variPtr;
          variPtr = &(np->variants[0]);
          *pval = variPtr[1];
        }
        return TRUE;
      }
    }
    return FALSE; /* not found */
  } else NULL_ARG_ERROR();
  return FALSE;
}

NLM_EXTERN void LIBCALL Nlm_VntMapDeleteAt (VntMapPtr vmp, VntPos vnp,
                                        VntProc finikey, VntProc finival)
{
  if (vmp != NULL && vnp != NULL) {
    vnode_ptr_t np = (vnode_ptr_t)vnp;
    if (finikey != NULL) finikey (&np->variants[0]);
    if (finival != NULL) finival (&np->variants[1]);
    vlist_remove ((vlist_ptr_t) vmp, np);
  } else NULL_ARG_ERROR();
}

NLM_EXTERN void LIBCALL Nlm_VntMapDeleteAll (VntMapPtr vmp, VntProc finikey, 
                                         VntProc finival)
{
  if (vmp != NULL) {
    vlist_ptr_t lp = (vlist_ptr_t) vmp;
    vnode_ptr_t np = lp->pfirst;
    while (np != NULL) {
      vnode_ptr_t pnext = np->pnext;
      if (finikey != NULL) finikey (&np->variants[0]);
      if (finival != NULL) finival (&np->variants[1]);
      free_vnode (np, lp);
      np = pnext;
    }
    lp->pfirst = lp->plast = NULL;
    lp->length = 0;
    /* shrink store? */
  } else NULL_ARG_ERROR();
}

NLM_EXTERN Boolean LIBCALL Nlm_VntMapAddNewAssoc (VntMapPtr vmp, Variant key, 
                                              VntCmpProc cmpkeys, Variant val)
{
  if (vmp != NULL && cmpkeys != NULL) {
    vlist_ptr_t lp = (vlist_ptr_t) vmp;
    vnode_ptr_t np;
    for (np = lp->pfirst; np != NULL; np = np->pnext) {
      Int2 ord = cmpkeys (&key, &np->variants[0]);
      if (ord == 0) return FALSE; /* better change old assoc for key! */
      if (ord < 0) break;
    }
    np = vlist_insert_new (lp, TRUE, np);  /* insert before */
    if (np != NULL) {
      Variant PNTR variPtr;
      variPtr = &(np->variants[0]);
      variPtr[0] = key;
      variPtr[1] = val;
      return TRUE;
    } else NO_MEM_WARNING();
  } else NULL_ARG_ERROR();
  return FALSE;
}


