/* $Id: ncbithr.c,v 5.13 1996/12/03 21:48:33 vakatov Exp $ */
/*****************************************************************************

    Name: ncbithr.c

    Description: Internal functions for Multi-thread API library

    Author: Sergei Shavirin

   ***************************************************************************

                          PUBLIC DOMAIN NOTICE
              National Center for Biotechnology Information

    This software/database is a "United States Government Work" under the
    terms of the United States Copyright Act.  It was written as part of
    the author's official duties as a United States Government employee
    and thus cannot be copyrighted.  This software/database is freely
    available to the public for use. The National Library of Medicine and
    the U.S. Government have not placed any restriction on its use or
    reproduction.

    Although all reasonable efforts have been taken to ensure the accuracy
    and reliability of the software and data, the NLM and the U.S.
    Government do not and cannot warrant the performance or results that
    may be obtained by using this software or data. The NLM and the U.S.
    Government disclaim all warranties, express or implied, including
    warranties of performance, merchantability or fitness for any
    particular purpose.

    Please cite the author in any work or product based on this material.

   ***************************************************************************

    Modification History:
           30 Apr 1996 - Shavirin Sergei  -  original written

    Bugs and restriction on use:

    Notes:
 * $Log: ncbithr.c,v $
 * Revision 5.13  1996/12/03  21:48:33  vakatov
 * Adopted for 32-bit MS-Windows DLLs
 *
 * Revision 5.12  1996/11/25  19:02:27  vakatov
 * Inherit error posting flags and log-file from the parent thread.
 * NlmMutexInit() now accepts pointer to mutex as an argument -- to
 * guarantee no MT-mess during the mutex initialization.
 * Added function NlmMutexLockEx() to lock yet initialized mutex;
 * it initialize the mutex before locking, if necessary.
 * Semaphor functions, NlmThreadKill(), NlmMutexUnlock(), NlmMutexDestroy():
 * got rid of the return value discrepancy between different platforms;
 * now return 0 on success, -1 on error on all platforms.
 * NlmMutexLock<Ex>():  protected from "self-deadlock" under Solaris/POSIX.
 *
 * Revision 5.11  1996/10/16  00:54:53  epstein
 * add NCBI_NOTHREADS_AVAIL flag to disable threading
 *
 * Revision 5.10  1996/09/04  19:37:32  epstein
 * workaround for SGI4
 *
 * Revision 5.9  1996/09/04  14:34:44  kans
 * trivial fixes to allow compilation under Symantec C++
 *
 * Revision 5.8  1996/08/22  14:07:53  shavirin
 * Finaly NlmThreadsAvailable return TRUE for WIN32
 *
 * Revision 5.7  1996/08/07  17:07:58  shavirin
 * Added function NlmCPUNumber(void), that returns number of CPU
 * on calling host
 *
 * Revision 5.6  1996/08/02  19:52:38  vakatov
 * Most of the thread-handling code has been protected from being
 * compiled on the non-thread platforms
 *
 * Revision 5.5  1996/08/02  15:59:38  shavirin
 * Added support for POSIX threads. Implemented semaphores and
 * Read-Write lock objects for POSIX threads through POSIX mutexes
 *
 * Revision 5.4  1996/07/16  19:58:50  vakatov
 * [WIN32]  NlmThreadSelf() now returns handler rather than pseudo-handler.
 * [SOLARIS...] [POSIX...]  init_exit_arr() protected by "mutex" from.
 * [POSIX] "mutex"-related stuff implemented.
 * NlmThreadMemFree() auxiliary function added.
 * The "thread destructors" calling order has been changed from FIFO to LIFO.
 * Added NlmThreadCreateEx() function to specify the thread destructor
 * from outside the thread.
 * The program name property is to be automatially inherited by the threads.
 *
 * Revision 5.3  1996/07/09  18:13:37  vakatov
 * Added a basic MT-support for POSIX and WIN32-SDK(Windows-95 and Windows-NT)
 * API -- the thread creation, joining and terminating. The thread joining
 * (waiting for the terminating) interface and behaviour has been slightly
 * changed.
 * Tested the lately introduced "OnExit"-features for the new platforms.
 *
 * Revision 5.2  1996/07/05  21:09:22  vakatov
 * Added NlmThreadAddOnExit() and NlmThreadRemoveOnExit() functions allowing
 * the user to specify a set of functions to be automatically executed on the
 * thread exiting stage. In addition, the execution of NlmThreadExit() function
 * is guaranteed in the case when the thread's main function exits with
 * the "return" operator, without calling the NlmThreadExit() explicitly
 *
 * The concerning stuff is also implemented(however, not tested yet) for
 * POSIX and WIN32-SDK(MSVC++ and Borland C++) threads
 *
 * Tests are added at the end of the file(preprocessor variable TEST_MODULE)
 * to provide easy testing of the new(and some of the old) NCBI thread features
 *

*****************************************************************************/

#include <ncbithri.h>
#include <ncbithr.h>

#ifndef NCBI_NOTHREADS_AVAIL
#if defined(SOLARIS_THREADS_AVAIL) || defined(POSIX_THREADS_AVAIL) || (defined(WIN32) && defined(_MT))
#define NCBI_THREADS_AVAIL
#ifdef WIN32
#define WIN32_THREADS_AVAIL
#endif
#endif

#else

#ifdef SOLARIS_THREADS_AVAIL
#undef SOLARIS_THREADS_AVAIL
#endif
#ifdef POSIX_THREADS_AVAIL
#undef POSIX_THREADS_AVAIL
#endif
#ifdef WIN32_THREADS_AVAIL
#undef WIN32_THREADS_AVAIL
#endif
#endif


/* exported in "corepriv.h" */
TNlmMutex corelibMutex;


/***********************************************************************
 *  INTERNAL
 ***********************************************************************/


/* POSIX implementation suppose that semaphore and RW objects
   must be implemented by user himself .... */

#if defined (POSIX_THREADS_AVAIL)
typedef struct psema_t {
	Int4               count;
	pthread_mutex_t    mutex;
	pthread_cond_t     cond;
} psema_t;

typedef struct rwlock_t {
	Int4               readers;
	pthread_mutex_t    mutex;
	pthread_cond_t     cond_r;
	pthread_cond_t     cond_w;
} rwlock_t;
#endif


/*
 *  Functions and data structures providing storage and execution of
 *  the "thread destructor functions"
 */
typedef struct
{
  TNlmThreadOnExit func;
  VoidPtr          arg;
}  TOnExitData;

typedef struct
{
  size_t      size;
  TOnExitData user[1];
}  TOnExitArray, PNTR TOnExitArrayPtr;


#if   defined (SOLARIS_THREADS_AVAIL)
static  thread_key_t thread_exit_key;
#elif defined (POSIX_THREADS_AVAIL)
static pthread_key_t thread_exit_key;
#elif defined (WIN32_THREADS_AVAIL)
#ifdef _MSC_VER
static __declspec( thread ) TOnExitArrayPtr thread_exit_arr = NULL;
static __declspec( thread ) HANDLE          thread_self     = NULL;
#else
static TOnExitArrayPtr      __thread        thread_exit_arr = NULL;
static HANDLE               __thread        thread_self     = NULL;
#endif
#else
static TOnExitArrayPtr                      thread_exit_arr = NULL;
#endif


static Boolean init_exit_arr( void )
{
  static Boolean initialized = FALSE;

  if ( initialized )
    return TRUE;

#if   defined (SOLARIS_THREADS_AVAIL)
  {
    static mutex_t key_lock;
    mutex_lock( &key_lock );
    if (!initialized  &&
        thr_keycreate(&thread_exit_key, NULL) == 0)
      initialized = TRUE;
    mutex_unlock( &key_lock );
  }
#elif defined (POSIX_THREADS_AVAIL)
  {
    static pthread_mutex_t key_lock;
    pthread_mutex_lock( &key_lock );
    if (!initialized  &&
        pthread_key_create(&thread_exit_key, NULL) == 0)
      initialized = TRUE;
    pthread_mutex_unlock( &key_lock );
  }
#elif defined (WIN32_THREADS_AVAIL)
  initialized = TRUE;
#else
  initialized = TRUE;
#endif

  return initialized;
}


static TOnExitArrayPtr get_exit_arr( void )
{
  if ( !init_exit_arr() )
    return NULL;

  {{
#if   defined (SOLARIS_THREADS_AVAIL)
  VoidPtr arr;
  if (thr_getspecific(thread_exit_key, &arr) != 0)
    return NULL;
  return (TOnExitArrayPtr)arr;
#elif defined (POSIX_THREADS_AVAIL)
  return (TOnExitArrayPtr)pthread_getspecific( thread_exit_key );
#elif defined (WIN32_THREADS_AVAIL)
  return thread_exit_arr;
#else
  return thread_exit_arr;
#endif
  }}
}


static void set_exit_arr(TOnExitArrayPtr arr_ptr)
{
#if   defined (SOLARIS_THREADS_AVAIL)
  if (    thr_setspecific(thread_exit_key, (VoidPtr)arr_ptr) != 0)
    abort();
#elif defined (POSIX_THREADS_AVAIL)
  if (pthread_setspecific(thread_exit_key, (VoidPtr)arr_ptr) != 0)
    abort();
#elif defined (WIN32_THREADS_AVAIL)
  thread_exit_arr = arr_ptr;
#else
  thread_exit_arr = arr_ptr;
#endif
}



/***********************************************************************
 *  EXTERNAL
 ***********************************************************************/


/* --------------------  NlmThreadAddOnExit  ----------------------
   Purpose:     Adds a function to be called on the thread termination
   Parameters:  The function "func" will be called with the "arg" argument
                when the thread is terminating;  then, the later registered
                "destroying function" will be called earlier (LIFO order)
   Returns:     TRUE on success...
   NOTE:        Must be called from inside the thread
  ------------------------------------------------------------------*/
NLM_EXTERN Boolean NlmThreadAddOnExit(TNlmThreadOnExit func, VoidPtr arg)
{
  TOnExitArrayPtr arr_ptr = get_exit_arr();
  size_t size = (arr_ptr == NULL) ? 0 : arr_ptr->size;

  size_t i = 0;
  while (i < size  &&  arr_ptr->user[i].func != NULL)
    i++;

  if (i == size)
    {
      TOnExitArrayPtr new_arr = (TOnExitArrayPtr)
        Nlm_MemNew(sizeof(TOnExitArray) + size * sizeof(TOnExitData));
      if (new_arr == NULL)
        return FALSE;

      new_arr->size = size + 1;
      if (size != 0)
        {
          Nlm_MemCpy(new_arr->user, arr_ptr->user,
                     size * sizeof(TOnExitData));
          Nlm_MemFree( arr_ptr );
        }
      arr_ptr = new_arr;
    }

  arr_ptr->user[i].func = func;
  arr_ptr->user[i].arg  = arg;
  set_exit_arr( arr_ptr );
  return TRUE;
}


/* --------------------  NlmThreadRemoveOnExit  -------------------
   Purpose:     Removes the thread's termination function(s)
   Description: This function removes all previosly registered(for this
                thread) terminating functions whose attributes
                matches both "func" and "arg"
   NOTE:        Must be called from inside the thread
  ------------------------------------------------------------------*/
NLM_EXTERN void NlmThreadRemoveOnExit(TNlmThreadOnExit func, VoidPtr arg)
{
  TOnExitArrayPtr arr_ptr = get_exit_arr();
  size_t size = (arr_ptr == NULL) ? 0 : arr_ptr->size;

  size_t i;
  for (i = 0;  i < size;  i++)
    {
      if (arr_ptr->user[i].func == func  &&
          arr_ptr->user[i].arg  == arg)
        {
          Nlm_MemMove(arr_ptr->user + i, arr_ptr->user + i + 1,
                      (size - i - 1) * sizeof(TOnExitData));
          arr_ptr->user[size-1].func = NULL;
        }
    }
}


NLM_EXTERN void NlmThreadMemFree(VoidPtr ptr)
{
  Nlm_MemFree( ptr );
}


/* --------------------  NlmThreadWrapper  -------------------------
   Purpose:     To wrap the starting thread main function
   Parameters:  The user's main thread function with args and flags
   Returns:     New thread handle. -- to be used to identify the thread.
   Description: This function creates new thread - independent process
                inside current process, that share most resources of the
                parent thread.
   NOTE:        This function must be used in NlmThreadCreate() only!
  ------------------------------------------------------------------*/

#ifdef NCBI_THREADS_AVAIL
typedef struct
{
  TNlmThreadStart  func;
  VoidPtr          arg;
  TNlmThreadOnExit exit_func;
  VoidPtr          exit_arg;
  CharPtr          progName;
  CharPtr          logName;
  ErrOpts          errOpts;
} TNlmThreadWrapperArg, PNTR TNlmThreadWrapperArgPtr;


static
#if defined(SOLARIS_THREADS_AVAIL) || defined(POSIX_THREADS_AVAIL)
VoidPtr
#elif defined(WIN32_THREADS_AVAIL)
DWORD WINAPI
#endif
NlmThreadWrapper(VoidPtr wrapper_arg)
{
  VoidPtr ret_value;
  TNlmThreadWrapperArgPtr user = (TNlmThreadWrapperArgPtr)wrapper_arg;
  TNlmThreadStart startFunc    = user->func;
  VoidPtr         startFuncArg = user->arg;

#ifdef WIN32_THREADS_AVAIL
  /* duplicate and store an actual(not a pseudo-) handle to the thread */
  if ( !DuplicateHandle(GetCurrentProcess(), GetCurrentThread(),
                        GetCurrentProcess(), &thread_self,
                        0, FALSE, DUPLICATE_SAME_ACCESS) )
    abort();
#endif

  /* register the exiting function specified by the thread parent, if any */
  if (user->exit_func != NULL  &&
      !NlmThreadAddOnExit(user->exit_func, user->exit_arg))
    abort();

  /* store program name */
  if (user->progName != NULL)
    Nlm_SetAppProperty("ProgramName",    (void *)user->progName);

  /* set the error posting environment */
  if (user->logName != NULL)
    {
      Nlm_ErrSetLog( user->logName );
      user->logName = (CharPtr)Nlm_MemFree( user->logName );
    }

  ErrRestoreOptions( &user->errOpts );

  Nlm_Free( wrapper_arg );

  /* run the user's thread function */
  ret_value = (*startFunc)( startFuncArg );

  /* the regular exit from the thread */
  NlmThreadExit( ret_value );

  /* should never reach here... */
  abort();
  return 0;
}
#endif  /* NCBI_THREADS_AVAIL */


/* --------------------  NlmThreadCreate  ---------------------------
   Purpose:     To create new thread
   Parameters:  theStartFunction - address of the thread function
                *arg             - arguments for the new thread
                flags            - attributes for the thread:
		                   THR_RUN (default) or THR_SUSPEND
   Returns:     New thread handle. Will be used to identify the thread.
   Description: This function creates new thread - independent process
                inside current process, that share most resources of the
                parent thread.
  ------------------------------------------------------------------*/
NLM_EXTERN TNlmThread NlmThreadCreate(
	       TNlmThreadStart theStartFunction,   /* address of the thread
						      function */
	       VoidPtr arg,         /* arguments for the new thread */
	       Int4    flags        /* attributes for the thread */
	   )
{
  return NlmThreadCreateEx(theStartFunction, arg, flags, NULL, NULL);
}


/* --------------------  NlmThreadCreateEx  -------------------------
   Purpose:     To create new thread and register its destructor
   Parameters:  "theStartFunction", "arg", "flags"  -- see NlmThreadCreate
                "exit_func", "exit_arg"  -- these two allow one to add a
                function to be called when the thread is exiting -- it's
                the only convenient way to do this from outside of thread
                (see also the NlmThreadAddOnExit() function to get know
                how to do it from the thread's inside).
   Returns:     New thread handle. Will be used to identify the thread.
   Description: This function creates new thread - independent process
                inside current process, that share most resources of the
                parent thread.
   NOTE:        It's guaranteed that the "exit_func" (if non-NULL) will
                be called after any other exiting function specified
                from inside the thread
  ------------------------------------------------------------------*/
NLM_EXTERN TNlmThread NlmThreadCreateEx(TNlmThreadStart  theStartFunction,
                             VoidPtr          arg,
                             Int4             flags,
                             TNlmThreadOnExit exit_func,
                             VoidPtr          exit_arg
                             )
{
  TNlmThread thread_handle = (TNlmThread)0;

#ifdef NCBI_THREADS_AVAIL
  /* initialize thread wrapper func */
  TNlmThreadWrapperArgPtr wrapper_data =
    Nlm_Calloc(1, sizeof(TNlmThreadWrapperArg));
  if (wrapper_data == NULL)
    return NULL;
  wrapper_data->func      = theStartFunction;
  wrapper_data->arg       = arg;
  wrapper_data->exit_func = exit_func;
  wrapper_data->exit_arg  = exit_arg;
  wrapper_data->progName  =
    Nlm_StringSave( (char *)Nlm_GetAppProperty( "ProgramName" ) );
  wrapper_data->logName   = Nlm_StringSave( Nlm_ErrGetLogfile() );
  ErrSaveOptions( &wrapper_data->errOpts );

  /* create new thread */
#if   defined(SOLARIS_THREADS_AVAIL)
  {{
  if(thr_create(NULL, 0, NlmThreadWrapper, (void *)wrapper_data,
                (long)flags, (thread_t *)&thread_handle) != 0)
    thread_handle = (TNlmThread)0;
  }}
#elif defined(POSIX_THREADS_AVAIL)
  {{
  pthread_attr_t attr;
  pthread_attr_init( &attr );
  if(pthread_create((pthread_t *)&thread_handle, &attr,
                    NlmThreadWrapper, (void *)wrapper_data) != 0)
    thread_handle = (TNlmThread)0;
  pthread_attr_destroy( &attr );
  }}

#elif defined(WIN32_THREADS_AVAIL)
  {{
    DWORD thread_id;
    thread_handle = (TNlmThread)CreateThread(NULL, 0,
                                 NlmThreadWrapper, (LPVOID)wrapper_data,
                                 flags, &thread_id);
  }}
#endif

  /* is OK? */
  if (thread_handle == (TNlmThread)0)
    {
#ifdef _DEBUG
      abort();
#endif
      wrapper_data->logName  = Nlm_MemFree( wrapper_data->logName  );
      wrapper_data->progName = Nlm_MemFree( wrapper_data->progName );
      Nlm_Free( wrapper_data );
    }
#endif  /* NCBI_THREADS_AVAIL */

  return thread_handle;
}


/* ----------------------  NlmThreadSelf  --------------------------
   Purpose:     To get the thread handle
   Parameters:  None
   Returns:     theThread - thread handle NULL if error
                an error
   Description: NlmThreadSelf() returns the handle of the calling thread
  -----------------------------------------------------------------*/

NLM_EXTERN TNlmThread NlmThreadSelf(void)
{
#if   defined(SOLARIS_THREADS_AVAIL)
  return (TNlmThread)thr_self();
#elif defined(POSIX_THREADS_AVAIL)
  return (TNlmThread)pthread_self();
#elif defined(WIN32_THREADS_AVAIL)
  return (TNlmThread)thread_self;
#elif defined(MUTEX_DLL)
  return (TNlmThread)Nlm_GetAppProcessID();
#else             /* default */
  return (TNlmThread)0;
#endif
}


/* ----------------------  NlmThreadJoin  --------------------------
   Purpose:     Wait for the thread termination
   Parameters:  INPUT:  "wait_for" -- thread handle to wait for termination
                OUTPUT: "status"   -- the thread's exit status
   Returns:     zero value on success;  non-zero value on error
   Description: Blocks the calling thread until the thread specified by
                wait_for terminates. The specified thread must be in the
                current process. If wait_for is (TNlmThread)0, then
                NlmThreadJoin() waits for ANY thread in the process to
                terminate.
   NOTE:        Multiple threads cannot wait for the same thread to
                terminate. One thread will return successfully and the
                others will fail.
  -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmThreadJoin(TNlmThread  wait_for, /* handle of the thread to wait */
                   VoidPtr    *status    /* the thread's exit status     */
                  )
{
#if   defined(SOLARIS_THREADS_AVAIL)
  return thr_join((thread_t)wait_for, NULL, status);

#elif defined(POSIX_THREADS_AVAIL)
  return pthread_join((pthread_t)wait_for, status);

#elif defined(WIN32_THREADS_AVAIL)
  HANDLE thread = (HANDLE)wait_for;
  WaitForSingleObject(thread, INFINITE);
  if (!GetExitCodeThread(thread, (LPDWORD)status)  ||
      (DWORD)*status == (DWORD)STILL_ACTIVE)
    {
      DWORD  err_code = GetLastError();
      if (err_code == 0)
        return -1;
      return (Int4)err_code;
    }

  CloseHandle( thread );
  return 0;

#else          /* default */
  return -1;
#endif
}

/* ----------------  NlmThreadSetConcurrency  ----------------------
   Purpose:     Set thread concurrency level
   Parameters:  new_level - concurrency level to set
   Returns:     0 is returned when successful. A non-zero value indicates
                an error
   Description: This function permits the application to give the threads
                system a hint, specified by new_level for the desired level
		of concurrency. The actual number of simultaneously active
		threads may be larger or smaller than this number.
   NOTE:        For Solaris Threads ONLY!
  -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmThreadSetConcurrency(Int4 new_level)
{
#if defined(SOLARIS_THREADS_AVAIL)
  return (Int4)thr_setconcurrency( new_level );

#else
  return 0;
#endif
}


/* ----------------  NlmThreadGetConcurrency  ----------------------
   Purpose:     To get thread concurrency level
   Parameters:  None
   Returns:     Current value for the desired concurrency level
   Description: returns the current value for the desired concurrency
                level. The actual number of simultaneously active threads
		may be larger or smaller than this number.
   NOTE:        For Solaris Threads ONLY!
  -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmThreadGetConcurrency(void)
{
#if defined(SOLARIS_THREADS_AVAIL)
  return (Int4)thr_getconcurrency();

#else
  return 0;
#endif
}


/* ----------------------  NlmThreadContinue  -----------------------
   Purpose:     Resume execution of a suspended thread
   Parameters:  theThread - thread handle to continue
   Returns:     0 is returned when successful. A non-zero value indicates
                an error
   Description: Resume execution of the thread, that was initially
                started in suspended state
   NOTE:        Currently suspending supported only during initial
                creation of the thread
  -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmThreadContinue(TNlmThread theThread)
{
#if defined(SOLARIS_THREADS_AVAIL)
  return (Int4)thr_continue( (thread_t)theThread );

#elif defined(WIN32_THREADS_AVAIL)
  if(ResumeThread( (HANDLE)theThread ) != 0xFFFFFFFF)
    return 0;
  else
    return 0xFFFFFFFF;

#else
  return 0;
#endif
}


/* -------------------  NlmThreadExit  ------------------------------
   Purpose:      To terminate the calling thread
   Parameters:   status - exit code for the calling thread
   Returns:      Nothing
   Description:  This function terminates th calling thread. All thread-
                 specific data bindings are released. If the initial thread
                 returns from main() then the process exits with a status
                 equal to the return value
   -----------------------------------------------------------------*/
NLM_EXTERN void NlmThreadExit(VoidPtr status  /* exit data of the calling thread */
                   )
{
  /* execute all terminating functions registered by user */
  TOnExitArrayPtr arr_ptr = get_exit_arr();
  size_t size = (arr_ptr == NULL) ? 0 : arr_ptr->size;
  while ( size-- )
    {
      if (arr_ptr->user[size].func != NULL)
        (*arr_ptr->user[size].func)( arr_ptr->user[size].arg );
    }

  if (arr_ptr != NULL)
    {
      Free( arr_ptr );
      set_exit_arr( (TOnExitArrayPtr)(-1) );
    }

#ifdef NCBI_THREADS_AVAIL
  /* exiting from thread must clear "context" structure */
  {{
  const char *progName = Nlm_GetAppProperty( "ProgramName" );
  Nlm_ReleaseAppContext();
  Nlm_MemFree( (void *)progName );
  }}

#if defined(SOLARIS_THREADS_AVAIL)
  thr_exit     ( (void *) status );
#elif defined(POSIX_THREADS_AVAIL)
  pthread_exit ( (void *) status );
#elif defined(WIN32_THREADS_AVAIL)
  CloseHandle( thread_self );
  ExitThread( (DWORD)  status );
#endif
#endif  /* NCBI_THREADS_AVAIL */
}


/* -------------------  NlmThreadSetPriority  -----------------------
   Purpose:     To set priority of the specified thread
   Parameters:  theThread - thread handle to set priority
                priority  - integer from 0 (the least priority)
                           to 127 (highest priority)
   Returns:     Zero is returned when succesful. A non-zero value
                indicates error.
   Description: Each thread has a priority which it inherits from its
                creator. This function changes priority of "theThread" to
                the value specified by "priority".
   NOTE:  On Solaris POSIX SCHED_OTHER policy implemented - use this so far
   -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmThreadSetPriority(TNlmThread theThread, /* thread handle */
                                     Int4 priority         /* priority level */
                                     )
{
#if defined(SOLARIS_THREADS_AVAIL)
  return (Int4)thr_setprio((thread_t)theThread, priority);

#elif defined (POSIX_THREADS_AVAIL)
  struct sched_param param;
  param.sched_priority = priority;
  return (Int4)pthread_setschedparam((pthread_t)theThread,
                                     SCHED_OTHER, &param);

#elif defined(WIN32_THREADS_AVAIL)
  return SetThreadPriority((HANDLE)theThread, priority) ? 0 : -1;

#else
  return 0;
#endif
}


/* -------------------  NlmThreadGetPriority  -----------------------
   Purpose:      To get priority of the thread
   Parameters:   theThread - handle of the thread to get priority
                 priority  - storage for the priority of theThread
   Returns:      Zero is returned, when succesful. A non-zero value
                 indicates error
   Description:  This function gets current value of priority of theThread
   NOTE: On Solaris POSIX only SCHED_OTHER policy implemented - use this so far
   -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmThreadGetPriority(TNlmThread theThread,
                                     int *priority)
{
#if defined(SOLARIS_THREADS_AVAIL)
  return thr_getprio((thread_t)theThread, priority);

#elif defined (POSIX_THREADS_AVAIL)
  struct sched_param param;
  int policy; /* SCHED_OTHER the only policy for Solaris */

  Int4 error = pthread_getschedparam((pthread_t)theThread, &policy, &param);
  if (error == 0)
    *priority = param.sched_priority;
  return error;

#elif defined(WIN32_THREADS_AVAIL)
  return GetThreadPriority( (HANDLE)theThread );

#else
  return 0;
#endif
}


/* ---------------------  NlmThreadKill  ----------------------------
   Purpose:     To terminate thread from another thread
   Parameters:  theThread - handle of the thread to kill
   Returns:     Zero is returned when successful. A non-zero value
                indicates an error
   Description: This function terminates thread, specified as parameter.
   NOTE:        In Solaris we send signal "9" to the thread.
   -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmThreadKill(TNlmThread theThread)
{
#if   defined(SOLARIS_THREADS_AVAIL)
  return thr_kill((thread_t)theThread, 9);

#elif defined(POSIX_THREADS_AVAIL)
  return pthread_kill((pthread_t)theThread, 9);

#elif defined(WIN32_THREADS_AVAIL)
  return TerminateThread((HANDLE)theThread, 1) ? 0 : -1;

#else
  return 0;
#endif
}


/********************************************************************/
/*                                                                  */
/* =========== Semaphore support for the thread API library ======= */
/*                                                                  */
/********************************************************************/


/* ---------------------  NlmSemaInit  ------------------------------
   Purpose:     To create new semaphore
   Parameters:  count - number of threads, that may use semaphore
                in parallel
   Returns:     Semaphore handle
   Description: To create new semaphore with "count" simultaneous accesses.
   NOTE:        This semathore can be used to synchronize threads in this
                process only
   -----------------------------------------------------------------*/
NLM_EXTERN TNlmSemaphore NlmSemaInit(Uint4 count)
{
#if defined(SOLARIS_THREADS_AVAIL)
  sema_t *theSemaphore = (sema_t *)Malloc(sizeof(sema_t));

  if(theSemaphore == NULL ||
     sema_init(theSemaphore, count, USYNC_THREAD, NULL) != 0)
    return NULL;
  else
   return (TNlmSemaphore)theSemaphore;

#elif defined(POSIX_THREADS_AVAIL)
  psema_t *theSemaphore;

  if ((theSemaphore = (psema_t *)Malloc(sizeof(psema_t))) == NULL)
	  return NULL;

  theSemaphore->count = count;

  if((pthread_mutex_init(&theSemaphore->mutex, NULL)) != 0)
	  return NULL;

  if((pthread_cond_init(&theSemaphore->cond, NULL)) != 0)
	  return NULL;

  return (TNlmSemaphore)theSemaphore;

#elif defined(WIN32_THREADS_AVAIL)
  HANDLE theSemaphore;

  if((theSemaphore =
      CreateSemaphore(NULL, count, MAX_SEMAPHORE_COUNT, NULL)) != NULL)
    return (TNlmSemaphore)theSemaphore;
  else
    return NULL;

#else
  return NULL;
#endif
}


/* ---------------------  NlmSemaDestroy  ---------------------------
   Purpose:     To destroy specified semaphore.
   Parameters:  theSemaphore - handle of the semaphore
   Returns:     Zero is returned when succesful. A non-zero value
                indicates an error
   Description: destroys any state associated with the semaphore pointed
                to by theSemaphore
   NOTE:        The space for storing the semaphore is NOT freed.
   -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmSemaDestroy(TNlmSemaphore theSemaphore)
{
#if defined(SOLARIS_THREADS_AVAIL)
  Int4 error = sema_destroy( (sema_t *)theSemaphore );
  Free( theSemaphore );
  return error;

#elif defined(POSIX_THREADS_AVAIL)
  psema_t *PSemaphore = (psema_t *)theSemaphore;

  if (pthread_mutex_destroy(&PSemaphore->mutex) != 0  ||
      pthread_cond_destroy (&PSemaphore->cond)  != 0)
    {
      Free( theSemaphore );
      return -1;
    }
  Free( theSemaphore );
  return 0;

#elif defined(WIN32_THREADS_AVAIL)
  return CloseHandle( (HANDLE)theSemaphore ) ? 0 : -1;

#else
  return 0;
#endif
}


/* ---------------------  NlmSemaWait  ------------------------------
   Purpose:      To wait until semaphore is not zero.
   Parameters:   theSemaphore - handle of the semaphore
   Returns:      Zero is returned when successful. A non-zero value
                 indicates an error
   Description:  Blocks the calling thread until the count in the
                 semaphore pointed by theSemaphore becomes greater than zero
                 and then atomically decrements it
   -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmSemaWait(TNlmSemaphore theSemaphore)
{
#if defined(SOLARIS_THREADS_AVAIL)
  return sema_wait((sema_t *)theSemaphore);

#elif defined(POSIX_THREADS_AVAIL)
  psema_t *PSemaphore = (psema_t *)theSemaphore;

  if((pthread_mutex_lock(&PSemaphore->mutex)) != 0)
	 return -1;

  if (PSemaphore->count > 0)
	  PSemaphore->count--;
  else {
	  do {
		  pthread_cond_wait(&PSemaphore->cond, &PSemaphore->mutex);
	  } while (PSemaphore->count == 0);

	  PSemaphore->count--;
  }

  if((pthread_mutex_unlock(&PSemaphore->mutex)) != 0)
	  return -1;
  return 0;

#elif defined(WIN32_THREADS_AVAIL)
  return
    (WaitForSingleObject((HANDLE)theSemaphore, INFINITE) == WAIT_OBJECT_0) ?
    0 : -1;

#else
  return 0;
#endif
}


/* ---------------------  NlmSemaTryWait  ---------------------------
   Purpose:      To poll semaphore status
   Parameters:   theSemaphore - handle of the semaphore to poll
   Returns:      Zero is returned when successful. A non-zero value
                 indicates an error
   Description:  atomically decrements the count in the semaphore
                 pointed to by theSemaphore if the count is greater
                 than zero. Otherwise it returns an error.
  -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmSemaTryWait(TNlmSemaphore theSemaphore)
{
#if defined(SOLARIS_THREADS_AVAIL)
  return sema_trywait( (sema_t *)theSemaphore );

#elif defined (POSIX_THREADS_AVAIL)
  psema_t *PSemaphore = (psema_t *)theSemaphore;

  if((pthread_mutex_lock(&PSemaphore->mutex)) != 0)
    return -1;

  if (PSemaphore->count > 0) {
    PSemaphore->count--;
    if((pthread_mutex_unlock(&PSemaphore->mutex)) != 0)
      return -1;
    return 0;
  } else {
    if((pthread_mutex_unlock(&PSemaphore->mutex)) != 0)
      return -1;
    return -1;
  }

#elif defined(WIN32_THREADS_AVAIL)
  return
    (WaitForSingleObject((HANDLE)theSemaphore, 0) == WAIT_OBJECT_0) ?
    0 : -1;

#else
  return 0;
#endif
}


/* ---------------------  NlmSemaPost  ------------------------------
   Purpose:    To increase the count of semaphore by one
   Parameters: theSemaphore - handle of the semaphore to post
   Returns:    Zero is returned when succesful. A non-zero value
               indicates an error
   Description:  atomically increments the count semaphore pointed to by
                 thSemaphore, one is unblocked.
   -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmSemaPost(TNlmSemaphore theSemaphore)
{
#if defined(SOLARIS_THREADS_AVAIL)
  return sema_post((sema_t *)theSemaphore);

#elif defined (POSIX_THREADS_AVAIL)
  psema_t *PSemaphore = (psema_t *)theSemaphore;

  if((pthread_mutex_lock(&PSemaphore->mutex)) != 0)
	 return -1;

  PSemaphore->count++;

  pthread_cond_signal(&PSemaphore->cond);

  if((pthread_mutex_unlock(&PSemaphore->mutex)) != 0)
    return -1;
  return 0;

#elif defined(WIN32_THREADS_AVAIL)
  return ReleaseSemaphore((HANDLE)theSemaphore, 1, NULL) ? 0 : -1;

#else
  return 0;
#endif
}


/********************************************************************/
/*                                                                  */
/* === Semaphore-like object to keep read/write syncronization ==== */
/*                                                                  */
/********************************************************************/

/* ---------------------  NlmRWlock  --------------------------------
   Purpose:     Initialize readers/writer lock
   Parameters:  None
   Returns:     Handle of the new RW - lock or NULL if error occour
   Description: Readers/writer locks must be initialized before use. This
                function initialize new lock.
   -----------------------------------------------------------------*/
NLM_EXTERN TNlmRWlock NlmRWinit(void)
{
#if defined(SOLARIS_THREADS_AVAIL)
  int status;
  rwlock_t *theRWlock;

  if((theRWlock = Malloc(sizeof(rwlock_t))) == NULL)
    return NULL;

  if ((status = rwlock_init(theRWlock, USYNC_THREAD, NULL)) != 0)
    return NULL;
  else
    return (TNlmRWlock *)theRWlock;

#elif defined(POSIX_THREADS_AVAIL)
  rwlock_t *theRWlock;

  if ((theRWlock = (rwlock_t *)Malloc(sizeof(rwlock_t))) == NULL)
    return NULL;

  theRWlock->readers = 0;

  if((pthread_mutex_init(&theRWlock->mutex, NULL)) != 0)
	  return NULL;

  if((pthread_cond_init(&theRWlock->cond_r, NULL)) != 0)
	  return NULL;

  if((pthread_cond_init(&theRWlock->cond_w, NULL)) != 0)
	  return NULL;

  return (TNlmRWlock)theRWlock;

#else
  return NULL;
#endif
}


/* ---------------------  NlmRWdestroy  -----------------------------
   Purpose:      To destroy RW lock
   Parameters:   theRWlock - handle of the RW lock to destroy
   Returns:      Zero is returned when succesful. A non-zero value
                 indicates an error
   Description:  Destroys any state associated with the RW lock
                 pointed to by theRWlock.
   NOTE:         The space for storing the RWlock is not freed.
  -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmRWdestroy(TNlmRWlock theRWlock)
{
#if defined(SOLARIS_THREADS_AVAIL)
  Int4 error = rwlock_destroy((rwlock_t *)theRWlock);
  Free(theRWlock);
  return error;

#elif defined (POSIX_THREADS_AVAIL)
  rwlock_t *RWlock = (rwlock_t *)theRWlock;

  if (pthread_mutex_destroy(&RWlock->mutex) != 0  ||
      pthread_cond_destroy(&RWlock->cond_r) != 0  ||
      pthread_cond_destroy(&RWlock->cond_w) != 0)
    {
      Free( theRWlock );
      return -1;
    }
  Free(theRWlock);
  return 0;

#else
  return 0;
#endif
}


/* ---------------------  NlmRWrdlock  ------------------------------
   Purpose:      To acquire a read lock
   Parameters:   theRWlock - handle of the RW lock
   Returns:      Zero is returned when succesful. A non-zero value
                 indicates an error
   Description:  Acquires a read lock on the readers/writer lock
                 pointed to by theRWlock. If the RW lock is already
                 locked for writing, the calling thread blocks until
                 the write lock is released.
   NOTE          More than one thread may hold a read lock on a RW lock at any
                 one time.
  -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmRWrdlock(TNlmRWlock theRWlock)
{
#if defined(SOLARIS_THREADS_AVAIL)
  return rw_rdlock((rwlock_t *)theRWlock);

#elif defined(POSIX_THREADS_AVAIL)
  rwlock_t *PRWlock = (rwlock_t *)theRWlock;

  if((pthread_mutex_lock(&PRWlock->mutex)) != 0)
    return -1;

  if (PRWlock->readers >= 0)
    PRWlock->readers++;
  else {
    do {
      pthread_cond_wait(&PRWlock->cond_r, &PRWlock->mutex);
    } while (PRWlock->readers != 0);

    PRWlock->readers++;
  }

  if((pthread_mutex_unlock(&PRWlock->mutex)) != 0)
    return -1;
  return 0;

#else
  return 0;
#endif
}


/* ---------------------  NlmRWwrlock  ------------------------------
   Purpose:      To acquire a write lock
   Parameters:   theRWlock - handle of the RW lock
   Returns:      Zero is returned when succesful. A non-zero value
                 indicates an error
   Description:  Acquires a write lock on the readers/writer lock
                 pointed to by theRWlock. If the RW lock is already
                 locked for reading or writing, the calling thread blocks
                 until the lock is released.
   NOTE:         Only one thread may hold a write lock on a RW lock at any
                 one time.
  -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmRWwrlock(TNlmRWlock theRWlock)
{
#if defined(SOLARIS_THREADS_AVAIL)
  return rw_wrlock((rwlock_t *)theRWlock);

#elif defined(POSIX_THREADS_AVAIL)
  rwlock_t *PRWlock = (rwlock_t *)theRWlock;

  if((pthread_mutex_lock(&PRWlock->mutex)) != 0)
	 return -1;

  if (PRWlock->readers == 0)
    PRWlock->readers--; /* -1 is special value for writer */
  else {
    do {
      pthread_cond_wait(&PRWlock->cond_w, &PRWlock->mutex);
    } while (PRWlock->readers != 0);
    
    PRWlock->readers--;
  }

  if((pthread_mutex_unlock(&PRWlock->mutex)) != 0)
    return -1;
  return 0;

#else
  return 0;
#endif
}


/* ---------------------  NlmRWunlock  ------------------------------
   Purpose:      To unlock the RW lock
   Parameters:   theRWlock - handle of the RW lock
   Returns:      Zero is returned when succesful. A non-zero value
                 indicates an error
   Description:  The RW lock must be locked and the calling thread must
                 hold the lock either for reading or for writing. If
                 any other threads are waiting for the RW lock to become
                 available, one of them is unblocked.
   NOTE:         If the calling thread does not hold the lock for either
                 reading or writing no error status is returned and the
                 behavior of the program is undefined.
   -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmRWunlock(TNlmRWlock theRWlock)
{
#if defined(SOLARIS_THREADS_AVAIL)
  return rw_unlock((rwlock_t *)theRWlock);

#elif defined (POSIX_THREADS_AVAIL)
  rwlock_t *RWlock = (rwlock_t *)theRWlock;

  if((pthread_mutex_lock(&RWlock->mutex)) != 0)
    return -1;

  if(RWlock->readers == -1) { /* Hmm.. this is writer unlock */
    RWlock->readers++;
    pthread_cond_broadcast(&RWlock->cond_r);
    pthread_cond_signal(&RWlock->cond_w);
  } else {
    RWlock->readers--;        /* Hmm.. this is reader unlock */
    if(!RWlock->readers)
      pthread_cond_signal(&RWlock->cond_w); /* if we clear access to writer*/
  }

  if((pthread_mutex_unlock(&RWlock->mutex)) != 0)
    return -1;
  return 0;

#else
  return 0;
#endif
}



/* ---------------------  NlmRWtryrdlock  ---------------------------
   Purpose:      To attempt to acquire read lock
   Parameters:   theRWlock - handle of the RW lock
   Returns:      Zero is returned when succesful. A non-zero value
                 indicates an error
   Description:  Attempts to acquire a read lock on the RW lock
                 pointed to by theRWlock. If the RW lock is already
                 locked for writing, it returnes an error. Otherwise
                 the read lock is acquired
   -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmRWtryrdlock(TNlmRWlock theRWlock)
{
#if defined(SOLARIS_THREADS_AVAIL)
  return rw_tryrdlock((rwlock_t *)theRWlock);

#elif defined(POSIX_THREADS_AVAIL)
  rwlock_t *PRWlock = (rwlock_t *)theRWlock;

  if((pthread_mutex_lock(&PRWlock->mutex)) != 0)
    return -1;

  if (PRWlock->readers >= 0)
    PRWlock->readers++;
  else
    return -1;

  if((pthread_mutex_unlock(&PRWlock->mutex)) != 0)
    return -1;
  return 0;

#else
  return 0;
#endif
}


/* ---------------------  NlmRWtrywrlock  ---------------------------
   Purpose:       To attempt to acquire write lock
   Parameters:    theRWlock - handle of the RW lock
   Returns:       Zero is returned when succesful. A non-zero value
                  indicates an error
   Description:   Attempts to acquire a write lock on the RW lock
                  pointed to by theRWlock. If the RW lock is already
                  locked for reading or writing, it returnes an error.
                  Otherwise the write lock is acquired
   -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmRWtrywrlock(TNlmRWlock theRWlock)
{
#if defined(SOLARIS_THREADS_AVAIL)

  return rw_trywrlock((rwlock_t *)theRWlock);

#elif defined(POSIX_THREADS_AVAIL)

  rwlock_t *PRWlock = (rwlock_t *)theRWlock;

  if((pthread_mutex_lock(&PRWlock->mutex)) != 0)
	 return -1;

  if (PRWlock->readers = 0)
    PRWlock->readers--;
  else
    return -1;

  if((pthread_mutex_unlock(&PRWlock->mutex)) != 0)
    return -1;
  return 0;

#else
  return 0;
#endif
}


/* ---------------------  NlmMutexInit  ------------------------------
   Purpose:     To create and initialize mutex, if necessary
   Parameters:  Pointer to the mutex handle
   Returns:     New(or old -- if the referenced mutex already exists)
                mutex handle
   Description: Create new mutex if it doesn's not exist(i.e. if
                *theMutexPtr == NULL); do nothing if it already exists.
   NOTE:        This mutex can be used to synchronize threads in this
                process only.
   -----------------------------------------------------------------*/
#if defined(SOLARIS_THREADS_AVAIL)  ||  defined(POSIX_THREADS_AVAIL)  ||  defined(WIN32_THREADS_AVAIL)  ||  defined(DLL_MUTEX)

typedef struct
{
#if   defined (SOLARIS_THREADS_AVAIL)
  mutex_t          mutex;
#elif defined (POSIX_THREADS_AVAIL)
  pthread_mutex_t  mutex;
#elif defined (WIN32_THREADS_AVAIL)
  HANDLE           mutex;
#endif
  TNlmThread owner;
  long       counter;
}
TNlmInternalMutex;

#else

#define NO_MUTEX
#define DUMMY_MUTEX 1

#endif


NLM_EXTERN TNlmMutex NlmMutexInit(TNlmMutexPtr theMutexPtr)
{
  if (theMutexPtr == NULL)
    {
      ASSERT_HARD ( FALSE );
      return NULL;
    }

  if (*theMutexPtr != NULL)
    return *theMutexPtr;

#if defined(SOLARIS_THREADS_AVAIL)
  {{
  static mutex_t init_lock;
  mutex_lock( &init_lock );

  if (*theMutexPtr == NULL)
    {
      TNlmInternalMutex *theMutex =
        (TNlmInternalMutex *) Calloc(1, sizeof(TNlmInternalMutex));
      if (theMutex != NULL  &&
          mutex_init(&theMutex->mutex, NULL, NULL) != 0)
        {
          Free( theMutex );
          theMutex = NULL;
        }
      *theMutexPtr = (TNlmMutex)theMutex;
    }

  mutex_unlock( &init_lock );
  }}

#elif defined(POSIX_THREADS_AVAIL)
  {{
  static pthread_mutex_t init_lock;
  pthread_mutex_lock( &init_lock );

  if (*theMutexPtr == NULL)
    {
      TNlmInternalMutex *theMutex =
        (TNlmInternalMutex *) Calloc(1, sizeof(TNlmInternalMutex));
      if (theMutex != NULL  &&
          pthread_mutex_init(&theMutex->mutex, NULL) != 0)
        {
          Free( theMutex );
          theMutex = NULL;
        }
      *theMutexPtr = (TNlmMutex)theMutex;
    }

  pthread_mutex_unlock( &init_lock );
  }}

#elif defined(WIN32_THREADS_AVAIL)
  {{
  static HANDLE init_lock = NULL;
  if (init_lock == NULL)
    init_lock = CreateMutex(NULL, FALSE, "Nlm_InitLock32");
  if (WaitForSingleObject(init_lock, INFINITE) != WAIT_OBJECT_0)
    return FALSE;

  if (*theMutexPtr == NULL)
    {
      TNlmInternalMutex *theMutex =
        (TNlmInternalMutex *) Calloc(1, sizeof(TNlmInternalMutex));
      if (theMutex != NULL  &&
          (theMutex->mutex = CreateMutex(NULL, FALSE, NULL)) == NULL)
        {
          Free( theMutex );
          theMutex = NULL;
        }
      *theMutexPtr = (TNlmMutex)theMutex;
    }

  ReleaseMutex( init_lock );
  }}

#elif defined(DLL_MUTEX)
  *theMutexPtr = Calloc(1, sizeof(TNlmInternalMutex));

#else
  *theMutexPtr = (TNlmMutex)DUMMY_MUTEX;
#endif

  return *theMutexPtr;
}


/* ---------------------  NlmMutexLock  ------------------------------
   Purpose:      To lock mutex.
   Parameters:   theMutex -- handle of the mutex
   Returns:      Zero is returned when successful. A non-zero value
                 indicates an error
   Description:  Blocks code access
   NOTE:
   -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmMutexLock(TNlmMutex theMutex)
{
#ifdef NO_MUTEX
  return 0;

#else /* NO_MUTEX */
  TNlmInternalMutex *_theMutex = (TNlmInternalMutex *)theMutex;
  TNlmThread         owner     = NlmThreadSelf();
  Int4               err_code  = 0;

  if (_theMutex == NULL)
    {
      ASSERT_HARD ( FALSE );
      return -1;
    }

  if (_theMutex->counter > 0  &&  _theMutex->owner == owner)
    { /* protect from deadlock under Solaris || POSIX */
      _theMutex->counter++;
      return 0;
    }

#if defined(_DEBUG_HARD)
  fprintf(stderr, "Mutex lock , IN = [%2ld :%8ld]\n",
          (long)owner, (long)theMutex);
#endif

#if defined(SOLARIS_THREADS_AVAIL)
  err_code = mutex_lock( &_theMutex->mutex );

#elif defined(POSIX_THREADS_AVAIL)
  err_code = pthread_mutex_lock( &_theMutex->mutex );

#elif defined(WIN32_THREADS_AVAIL)
  if (WaitForSingleObject(_theMutex->mutex, INFINITE) != WAIT_OBJECT_0)
    err_code = -1;
#endif

#if defined(_DEBUG_HARD)
  fprintf(stderr, "Mutex lock ,OUT = [%2ld :%8ld]\n",
          (long)owner, (long)theMutex);
#endif

  if (err_code == 0)
    {
      _theMutex->owner   = owner;
      _theMutex->counter = 1;
    }

  return err_code;
#endif /* else!NO_MUTEX */
}


/* ---------------------  NlmMutexLockEx  ----------------------------
   Purpose:      To lock mutex(initialize it, if necessary)
   Parameters:   theMutex -- handle of the mutex
   Returns:      Zero is returned when successful. A non-zero value
                 indicates an error
   Description:  Blocks code access
   NOTE:
   -----------------------------------------------------------------*/
Int4 NlmMutexLockEx(TNlmMutexPtr theMutexPtr)
{
#ifdef NO_MUTEX
  return 0;

#else  /* NO_MUTEX */
  if (theMutexPtr == NULL  ||
      *theMutexPtr == NULL  &&  !NlmMutexInit(theMutexPtr))
    {
      ASSERT_HARD ( FALSE );
      return -1;
    }

  return NlmMutexLock( *theMutexPtr );
#endif  /* else!NO_MUTEX */
}


/* ---------------------  NlmMutexUnlock  ------------------------------
   Purpose:      To unlock mutex.
   Parameters:   theMutex - handle of the mutex
   Returns:      Zero is returned when successful. A non-zero value
                 indicates an error
   Description:  Unblocks code access
   NOTE:
   -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmMutexUnlock(TNlmMutex theMutex)
{
#ifdef NO_MUTEX
  return 0;

#else  /* NO_MUTEX */
  TNlmInternalMutex *_theMutex = (TNlmInternalMutex *)theMutex;
  Int4               err_code  = 0;

  if (_theMutex == NULL)
    {
#ifdef _DEBUG_HARD
      abort();
#endif
      return 0;
    }

  if (_theMutex->counter == 0)
    return 0;

  if (_theMutex->owner != NlmThreadSelf())
    return -1;

  if (--_theMutex->counter != 0)
    return 0;

#if defined(SOLARIS_THREADS_AVAIL)
  err_code = mutex_unlock( &_theMutex->mutex );
#elif defined(POSIX_THREADS_AVAIL)
  err_code = pthread_mutex_unlock( &_theMutex->mutex );
#elif defined(WIN32_THREADS_AVAIL)
  err_code = ReleaseMutex( _theMutex->mutex ) ?  0 : GetLastError();
#endif

#if defined(_DEBUG_HARD)
  fprintf(stderr, "Mutex unlock ,  = [%2ld :%8ld / %3ld] :: %10ld\n",
          (long)NlmThreadSelf(), (long)theMutex, (long)_theMutex->counter,
          (long)err_code);
#endif

  if (err_code != 0)
    _theMutex->counter++;

  return err_code;
#endif  /* else!NO_MUTEX */
}


/* ---------------------  NlmMutexTryLock  ------------------------------
   Purpose:      To poll mutex.
   Parameters:   theMutex -- handle of the mutex
   Returns:      Zero is returned when successful. A non-zero value
                 indicates an error
   Description:  Poll condition of mutex -- do not block if mutex busy
   NOTE:
   -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmMutexTryLock(TNlmMutex theMutex)
{
#ifdef NO_MUTEX
  return 0;

#else  /* NO_MUTEX */
  TNlmInternalMutex *_theMutex = (TNlmInternalMutex *)theMutex;

  if (_theMutex == NULL)
    {
#ifdef _DEBUG_HARD
      abort();
#endif
      return -1;
    }

  if (_theMutex->counter == 0)
    return NlmMutexLock( theMutex );

  if (_theMutex->owner != NlmThreadSelf())
    return -2;

  _theMutex->counter++;
  return 0;
#endif  /* else!NO_MUTEX */
}


/* ---------------------  NlmMutexDestroy ------------------------------
   Purpose:      To destroy mutex.
   Parameters:   theMutex - handle of the mutex
   Returns:      Zero is returned when successful. A non-zero value
                 indicates an error
   Description:  Remove the mutex from execution.
   NOTE:
   -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmMutexDestroy(TNlmMutex theMutex)
{
#ifdef NO_MUTEX
  return 0;

#else  /* NO_MUTEX */
  TNlmInternalMutex *_theMutex = (TNlmInternalMutex *)theMutex;
  Int4               err_code  = 0;

  if (_theMutex == NULL)
    {
#ifdef _DEBUG_HARD
      abort();
#endif
      return 0;
    }

#if defined(SOLARIS_THREADS_AVAIL)
  err_code = mutex_destroy( &_theMutex->mutex );

#elif defined(POSIX_THREADS_AVAIL)
  err_code = pthread_mutex_destroy( &_theMutex->mutex );

#elif defined(WIN32_THREADS_AVAIL)
  if ( !CloseHandle( _theMutex->mutex ) )
    err_code = GetLastError();
#else
#endif

  Free( _theMutex );
  return err_code;
#endif  /* else!NO_MUTEX */
}



/* ---------------------  NlmCPUNumber ------------------------------
   Purpose:      To get number of CPU on machine
   Parameters:   none
   Returns:      Number of CPU
   -----------------------------------------------------------------*/
NLM_EXTERN Int4 NlmCPUNumber(void)
{
#if defined(WIN32)
  SYSTEM_INFO sysInfo;
  GetSystemInfo(&sysInfo);
  return sysInfo.dwNumberOfProcessors;
#elif defined(OS_UNIX_SYSV) && defined(SYSPROC_SPARC)  /* SPARC Solaris */
  return  sysconf(_SC_NPROCESSORS_ONLN);
#elif defined(OS_UNIX_SYSV) && defined(PROC_I80X86)  /* Intel Solaris */
  return  sysconf(_SC_NPROCESSORS_ONLN);
#elif  defined(PROC_MIPS) && defined(_SC_NPROC_ONLN)
  return sysconf(_SC_NPROC_ONLN);
#else
  return -1;
#endif
}


/* ---------------------  NlmThreadsAvailable ------------------------------
   Purpose:      Check to see if threads are available on this platform
   Parameters:   none
   Returns:      TRUE if threads are available, FALSE otherwise
   -----------------------------------------------------------------*/
NLM_EXTERN Boolean NlmThreadsAvailable(void)
{
#ifdef NCBI_THREADS_AVAIL
  return TRUE;
#else
  return FALSE;
#endif
}



#ifdef TEST_MODULE_NCBITHR
/***********************************************************************
 *  TEST
 ***********************************************************************/

#ifndef NCBI_THREADS_AVAIL

!!!!! ATTENTION -- BAD PLATFORM TO TEST THREADS  or  THREADS DISABLED !!!!!

#endif


#include <stdio.h>

#ifdef WIN32
#define STDERR stderr_file
static FILE *STDERR = NULL;
#else
#define STDERR stderr
#endif

#define DO_POSTEX
#define DO_STDERR

static void TEST__ThreadExit0(VoidPtr user_arg)
{
#ifdef DO_STDERR
  fprintf(STDERR, "ThreadExit0():  Self=%ld, Arg = %p\n",
          (long)NlmThreadSelf(), user_arg);
  Nlm_MemFree( user_arg );
#endif
}

static void TEST__ThreadExit1(VoidPtr user_arg)
{
#ifdef DO_STDERR
  fprintf(STDERR, "ThreadExit1():  Self=%ld, Arg = %ld\n",
          (long)NlmThreadSelf(), (long)user_arg);
#endif
}

static void TEST__ThreadExit2(VoidPtr user_arg)
{
#ifdef DO_STDERR
  fprintf(STDERR, "ThreadExit2():  Self=%ld, Arg = %ld\n",
          (long)NlmThreadSelf(), (long)user_arg);
#endif
}


TNlmThreadOnExit exit_func_arr[] =
{
  TEST__ThreadExit0,
  TEST__ThreadExit1,
  TEST__ThreadExit2,
  TEST__ThreadExit1,
  TEST__ThreadExit2,
  NULL
};

typedef struct
{
  Int4                  n_exit_func;
  TNlmThreadOnExit PNTR exit_func;
}  MyExitSet, PNTR MyExitSetPtr;


static Int2 TEST__threads(Int4 n_threads);

static VoidPtr TEST__MyThread(VoidPtr arg)
{
  MyExitSetPtr exit_func_set = (MyExitSetPtr) arg;
  Int4 f;
#ifdef DO_STDERR
  fprintf(STDERR, "TEST__MyThread():  Thread #%ld started\n",
          (long)NlmThreadSelf());
#endif
#ifdef DO_POSTEX
  Nlm_ErrPostEx(SEV_INFO, 2, 3, "TEST__MyThread():  Thread #%ld started\n",
          (long)NlmThreadSelf());
#endif

  {{
#if defined(SOLARIS_THREADS_AVAIL) || defined(POSIX_THREADS_AVAIL)
#define NCBI_MAX_THREADS _POSIX_THREAD_THREADS_MAX
#elif defined(WIN32)
#define NCBI_MAX_THREADS 32
#else
#define NCBI_MAX_THREADS INT2_MAX
#endif

#define RESERVE_THREADS 8
#define N_SPAWN         3


    static int nnn_threads = RESERVE_THREADS + N_SPAWN;
    static TNlmMutex xxx;
    int more;
    NlmMutexLockEx( &xxx );
    more = (nnn_threads < NCBI_MAX_THREADS - N_SPAWN);
    if ( more )
      nnn_threads += N_SPAWN;
    NlmMutexUnlock( xxx );
    if ( more )
      TEST__threads( N_SPAWN );
  }}

  /* register the dummy(debug-print only) exit functions */
  for (f = exit_func_set->n_exit_func - 1;  f > 0;  f--)
    {
      NlmThreadAddOnExit(exit_func_set->exit_func[f],
                         (VoidPtr)NlmThreadSelf());

      if (f%2 == 0)
        NlmThreadRemoveOnExit(exit_func_set->exit_func[f],
                              (VoidPtr)NlmThreadSelf());
    }

  /* register the exit function to deallocate dynamic memory */
  NlmThreadAddOnExit(exit_func_set->exit_func[0], (VoidPtr)exit_func_set);

  /* sleep for a while before exiting */
  {{
    static int do_sleep = 0;
    if ((do_sleep = !do_sleep) == TRUE)
      {
#if defined (OS_UNIX)
        sleep( 1 );
#elif defined (WIN32)
        Sleep( 1000 );
#endif
      }
  }}

#ifdef DO_POSTEX
  Nlm_ErrPostEx(SEV_INFO, 2, 3, "TEST__MyThread():  Thread #%ld finished\n",
          (long)NlmThreadSelf());
#endif

  return NULL;
}


static Int2 TEST__threads(Int4 n_threads)
{
  Int4 t;
  TNlmThread *threads = Nlm_MemNew(n_threads * sizeof(TNlmThread));

  for (t = 0;  t < n_threads;  t++)
    {
      MyExitSetPtr exit_set = (MyExitSetPtr) Nlm_MemNew( sizeof(MyExitSet) );
      exit_set->n_exit_func = t + 1;
      exit_set->exit_func   = exit_func_arr;
      threads[t] = NlmThreadCreate(&TEST__MyThread, (VoidPtr)exit_set, THR_RUN);
#ifdef DO_STDERR
      fprintf(STDERR, "TEST__threads():  Starting thread #%ld\n",
              (long)threads[t]);
#endif
    }

  for (t = 0;  t < n_threads;  t++)
    {
      VoidPtr status;
      Int4 err_code = NlmThreadJoin(threads[t], &status);
#ifdef DO_STDERR
      if (err_code != 0)
        fprintf(STDERR, "TEST__threads():  Cannot join thread #%ld;  \
error code = %ld\n",
                (long)threads[t], (long)err_code);

      fprintf(STDERR, "TEST__threads():  Thread #%ld joined;  \
terminated, exit status = %p\n",
              (long)threads[t], status);
#endif
    }

  Nlm_MemFree( threads );

#ifdef DO_STDERR
  fprintf(STDERR, "TEST__threads():  FINISHED\n");
#endif
  return 0;
}


Nlm_Int2 Nlm_Main( void )
{
#if defined (WIN32)
  STDERR = fopen("stderr", "w");
#endif

  {{
  int ok = Nlm_ErrSetLog( "ncbithr.log" );
  ASSERT ( ok );
  Nlm_ErrSetOpts(ERR_TEE, ERR_LOG_ON);
  Nlm_ErrPostEx(SEV_INFO, 2, 3, "NCBI_THREAD test started...\n");
  }}

  return TEST__threads( N_SPAWN );
}

#endif  /* TEST_MODULE_NCBITHR */

/* EOF */
